from .file import (
    check_file_overwrite,
    get_file_path,
    get_file_extension,
    get_parent_path,
    get_abs_path,
    get_real_path,
    is_binary_file,
    is_file,
    is_directory,
    is_json,
    open_json,
    open_file,
    open_yaml_file,
    write_file,
    write_json,
    write_yaml,
    get_script_path

)
from .color_print import (
    bcolors,
    colored,
    cprint,
    print_here,
    get_debug_here_info,
    get_variable_name_list,
    get_variable_name,
    dict_clean,
    list_clean,
    data_clean,
    count_nested_dict_len,
    print_frames,
    PrintRichTable,
    TablePrinter,
    get_bcolors,
    colored_input,
    dump,
    debug_print,
    classdump,
    kvPrint,
    print_json,
    print_syntax,
    pretty_json,
    print_aligned_text,
    print_var,
    debug_logging,
    align_text,
    print_progress_bar,
    get_colorful_object,
    syntax_highlight,
    ProgressTime,
    NoTraceBackException
)



class Null(object):
    """
    A Null object class as part of the Null object design pattern.
    """

    def __init__(self, *args, **kwargs):
        """
        Do nothing.
        """
        pass

    def __call__(self, *args, **kwargs):
        """
        Do nothing.
        @return: This object instance.
        @rtype: Null
        """
        return self

    def __getattr__(self, name):
        """
        Do nothing.
        @return: This object instance.
        @rtype: Null
        """
        return self

    def __setattr__(self, name, value):
        """
        Do nothing.
        @return: This object instance.
        @rtype: Null
        """
        return self

    def __delattr__(self, name):
        """
        Do nothing.
        @return: This object instance.
        @rtype: Null
        """
        return self

    def __repr__(self):
        """
        Null object string representation is the empty string.
        @return: An empty string.
        @rtype: String
        """
        return ''

    def __str__(self):
        """
        Null object string representation is the empty string.
        @return: An empty string.
        @rtype: String
        """
        return ''

    def __bool__(self):
        """
        Null object evaluates to False.
        @return: False.
        @rtype: Boolean
        """
        return False



from pawnlib.config import pawn, pconf


class _AttributeHolder(object):
    """Abstract base class that provides __repr__.

    The __repr__ method returns a string in the format::
        ClassName(attr=name, attr=name, ...)
    The attributes are determined either by a class-level attribute,
    '_kwarg_names', or by inspecting the instance __dict__.
    """

    def __repr__(self):
        type_name = type(self).__name__
        arg_strings = []
        star_args = {}
        for arg in self._get_args():
            arg_strings.append(repr(arg))
        for name, value in self._get_kwargs():
            if name.isidentifier():
                arg_strings.append('%s=%r' % (name, value))
            else:
                star_args[name] = value
        if star_args:
            arg_strings.append('**%s' % repr(star_args))
        return '%s(%s)' % (type_name, ', '.join(arg_strings))

    def _get_kwargs(self):
        return sorted(self.__dict__.items())

    def _get_args(self):
        return []

    def _set_args(self, key, value):
        self.__dict__[key] = value

    def _update(self, **kwargs):
        self.__dict__.update(kwargs)


class Namespace(_AttributeHolder):
    """Simple object for storing attributes.

    Implements equality by attribute names and values, and provides a simple
    string representation.

    Example:

        .. code-block:: python

            from pawnlib.typing import defines

            namespace = defines.Namespace(s=2323, sdsd="Sdsd")
            namespace.s
            # >> 2323
            namespace.sdsd
            # >> 'Sdsd'
    """

    def __init__(self, **kwargs):
        for name in kwargs:
            setattr(self, name, kwargs[name])

    def __eq__(self, other):
        if not isinstance(other, Namespace):
            return NotImplemented
        return vars(self) == vars(other)

    def __contains__(self, key):
        return key in self.__dict__


def set_namespace_default_value(namespace=None, key='', default=""):
    """
    Set a default value when a key in a namespace has no value

    :param namespace:
    :param key:
    :param default:
    :return:

    Example:

        .. code-block:: python

            from pawnlib.config import pawn, pconf
            from pawnlib.typing import set_namespace_default_value

            pawn.set(
            data={"aaaa": "bbbb"}
            )
            pawn.console.log(pconf())
            undefined_key = set_namespace_default_value(
                namespace=pconf().data,
                key="cccc",
                default="ddddd"
            )
            pawn.console.log(undefined_key)

    """
    if key and hasattr(namespace, key):
        return getattr(namespace, key)
    return default


def fill_required_data_arguments(required={}):
    """
     Fill the required data arguments.

     :param required: A dictionary of required arguments.
     :type required: dict
     :return: The filled arguments.
     :rtype: argparse.Namespace

     Example:

         .. code-block:: python

             required = {"arg1": "value1", "arg2": "value2"}
             args = fill_required_data_arguments(required)
             # args.arg1 == "value1"
             # args.arg2 == "value2"

     """
    none_string = "__NOT_DEFINED_VALUE__"
    if getattr(pconf(), "data", None) and getattr(pconf().data, "args", None):
        args = pconf().data.args
        for req_key, req_value in required.items():
            args_value = getattr(args, req_key, none_string)
            if args_value == none_string:
                # pawn.console.debug(f"Define the data args -> {req_key}, {req_value}")
                setattr(args, req_key, req_value)
    else:
        # pawn.console.debug(f"New definition: {required}")
        args = Namespace(**required)
    return args



#!/usr/bin/env python3
import argparse
from pawnlib.builder.generator import generate_banner
from pawnlib.__version__ import __version__ as _version
from pawnlib.config import pawn, pconf
import json
import copy
from pawnlib.typing import str2bool, StackList, is_json, is_valid_url, sys_exit, Null
from pawnlib.utils.http import CallHttp, disable_ssl_warnings
from pawnlib.utils import ThreadPoolRunner, send_slack
from pawnlib.output import get_script_path, dump, debug_print, print_var
from pawnlib.typing import remove_tags
import os

__description__ = 'This is a tool to measure RTT on HTTP/S requests.'


def get_parser():
    parser = argparse.ArgumentParser(description='httping')
    parser = get_arguments(parser)
    return parser


def get_arguments(parser):
    parser.add_argument('url', help='url', type=str, nargs='?', default="")
    parser.add_argument('-c', '--config-file', type=str, help='config', default="config.ini")
    parser.add_argument('-v', '--verbose', action='count', help='verbose mode. view level (default: %(default)s)', default=1)
    parser.add_argument('-q', '--quiet', action='count', help='Quiet mode. Dont show any messages. (default: %(default)s)', default=0)
    parser.add_argument('-i', '--interval', type=float, help='interval sleep time seconds. (default: %(default)s)', default=1)
    parser.add_argument('-m', '--method', type=lambda s : s.upper(), help='method. (default: %(default)s)', default="get")
    parser.add_argument('-t', '--timeout', type=float, help='timeout seconds (default: %(default)s)', default=10)
    parser.add_argument('-b', '--base-dir', type=str, help='base dir for httping (default: %(default)s)', default=os.getcwd())
    parser.add_argument('--success', nargs='+', help='success criteria. (default: %(default)s)', default=['status_code==200'])
    parser.add_argument('--logical-operator',
                        type=str,
                        help='logical operator for checking success condition (default: %(default)s)',
                        choices=["and", "or"],
                        default="and"
                        )
    parser.add_argument('--ignore-ssl', type=str2bool, help='ignore ssl certificate (default: %(default)s)', default=True)
    parser.add_argument('-d', '--data', type=json.loads, help="data parameter", default={})
    parser.add_argument('--headers', type=json.loads, help="header parameter", default={})
    parser.add_argument('-w', '--workers', type=int, help="max worker process (default: %(default)s)", default=10)
    parser.add_argument('--stack-limit', type=int, help="error stack limit (default: %(default)s)", default=3)
    parser.add_argument('--dynamic-increase-stack-limit', type=str2bool, help="error stack limit (default: %(default)s)", default=1)
    parser.add_argument('--slack-url', type=str, help="Slack URL", default="")
    return parser


def check_url_process(args):
    if not args.url:
        return

    check_url = CallHttp(
        url=args.url,
        method=args.method,
        timeout=args.timeout * 1000,
        data=args.data,
        headers=args.headers,
        success_criteria=args.success,
        success_operator=args.logical_operator,
        ignore_ssl=args.ignore_ssl,
    ).run()
    args.total_count += 1

    if args.verbose == 0:
        check_url.response.text = ""

    response_time = check_url.response.elapsed
    args.response_time.push(response_time)
    avg_response_time = f"{int(args.response_time.mean())}"
    max_response_time = f"{int(args.response_time.max())}"
    min_response_time = f"{int(args.response_time.min())}"
    status_code = check_url.response.status_code

    if args.fail_count > 0:
        count_msg = f'CER:{args.error_stack_count}/[red]ER:{args.fail_count}[/red]/SQ:{args.total_count}'
    else:
        count_msg = f'CER:{args.error_stack_count}/ER:{args.fail_count}/SQ:{args.total_count}'

    message = f"<{count_msg}> name={args.section_name}, url={args.url}, " \
              f"status={status_code}, {response_time:>4}ms " \
              f"(avg: {avg_response_time}, max: {max_response_time}, min: {min_response_time})"

    if args.verbose > 0:
        if status_code != 999:
            if args.verbose > 2:
                detail = f" 📄[i]{check_url.response}, criteria: {args.success}, operator: {args.logical_operator}[/i]"
            else:
                detail = ""
            message = f"{message}{detail}"
        else:
            message = f"{message} 😞 "

    if check_url.is_success():
        pawn.app_logger.info(remove_tags(f"[ OK ] {message}"))
    else:
        handle_failure_on_check_url(args, message, check_url)


def handle_failure_on_check_url(args, message, check_url):
    args.fail_count += 1
    args.error_stack_count += 1

    if args.error_stack_count >= args.stack_limit:
        pawn.error_logger.error(remove_tags(f"[FAIL][OVERFLOW]{args.error_stack_count}/{args.stack_limit} "
                                            f"Error Stack Count: {args.error_stack_count}, SEND_SLACK"))
        args.error_stack_count = 0

        if args.dynamic_increase_stack_limit:
            args.stack_limit = args.stack_limit ** 2
            _send_slack(url=args.slack_url, title=f"Error {args.url}", msg_text=args.__dict__)

    pawn.error_logger.error(remove_tags(f"[FAIL] {message}, Error={check_url.response}"))


def set_default_counter(section_name="default"):
    args = copy.deepcopy(pconf().args)
    args.section_name = section_name
    args.response_time = StackList()
    args.error_stack_count = 0
    args.total_count = 0
    args.fail_count = 0
    return args


def generate_task_from_config():
    tasks = []
    pconfig = pconf().PAWN_CONFIG
    config_file = pconf().PAWN_CONFIG_FILE
    pconf_dict = pconfig.as_dict()
    if pconf_dict:
        pawn.console.log(f"Found config file={config_file}")
        for section_name, section_value in pconf_dict.items():
            pawn.console.debug(f"section_name={section_name}, value={section_value}")
            args = set_default_counter(section_name)

            for conf_key, conf_value in section_value.items():
                if getattr(args, conf_key, "__NOT_DEFINED__") != "__NOT_DEFINED__":
                    if is_json(conf_value):
                        conf_value = json.loads(conf_value)
                    if section_name == "default":
                        setattr(pconf().args, conf_key, conf_value)
                    pawn.console.debug(f"Update argument from {config_file}, <{section_name}> {conf_key}={conf_value} <ignore args={getattr(args, conf_key, None)}>")
                    setattr(args, conf_key, conf_value)

            pawn.console.debug(args)
            if args.url != "http":
                tasks.append(args)

    if not tasks:
        args = set_default_counter()
        tasks = [args]

    validate_task_exit_on_failure(tasks)
    return tasks


def _send_slack(url, msg_text, title=None, send_user_name=None, msg_level='info'):
    if not send_user_name:
        send_user_name = pconf().app_name

    if url:
        pawn.console.log("Send slack")
        return send_slack(
            url=url,
            msg_text=msg_text,
            title=title,
            send_user_name=send_user_name,
            msg_level=msg_level
        )


def print_banner():
    banner = generate_banner(
        app_name=pconf().app_name,
        author="jinwoo",
        description=f"{__description__} \n"
                    f" - base_dir    : {pconf().args.base_dir} \n" 
                    f" - logs_dir    : {pconf().args.base_dir}/logs \n",
        font="graffiti",
        version=_version
    )
    print(banner)


def validate_task_exit_on_failure(tasks):
    if not tasks:
        sys_exit("Task not found")
    is_least_one_url = False
    for task in tasks:
        if is_valid_url(task.url):
            is_least_one_url = True
        else:
            pawn.console.log(f"Invalid url: name={task.section_name}, url={task.url}")

    if not is_least_one_url:
        sys_exit("Requires at least one valid URL. The URL argument must be in the first position.")


def main():
    app_name = 'httping'
    parser = get_parser()
    args, unknown = parser.parse_known_args()
    config_file = args.config_file

    is_hide_line_number = args.verbose > 1
    stdout = not args.quiet

    pawn.set(
        PAWN_CONFIG_FILE=config_file,
        PAWN_PATH=args.base_dir,
        PAWN_LOGGER=dict(
            log_level="INFO",
            stdout_level="INFO",
            log_path=f"{args.base_dir}/logs",
            stdout=stdout,
            use_hook_exception=True,
            show_path=False, #hide line numbers
        ),
        PAWN_CONSOLE=dict(
            redirect=True,
            record=True,
            log_path=is_hide_line_number, # hide line number on the right side
        ),
        app_name=app_name,
        args=args,
        try_pass=False,
        last_execute_point=0,
        data={
            "response_time": StackList(),
        },
        fail_count=0,
        total_count=0,

    )
    if args.verbose > 2:
        pawn.set(
            PAWN_LOGGER=dict(
                log_level="DEBUG",
                stdout_level="DEBUG",
            )
        )
    print_banner()

    if args.ignore_ssl:
        disable_ssl_warnings()
    tasks = generate_task_from_config()
    # if args.slack_url:
    #     res = _send_slack(url=args.slack_url, msg_text=tasks)
    #     pawn.console.log(res)
    # _send_slack(url=args.slack_url, title=f"Error HTTPING {args.url}", msg_text=args.__dict__)
    # pawn.console.log(f"console_options={pawn.console_options}")
    # exit()
    pawn.console.log(f"Start httping ... url_count={len(tasks)}")
    pawn.console.log("If you want to see more logs, use the [yellow]-v[/yellow] option")
    pawn.console.log(tasks)

    ThreadPoolRunner(
        func=check_url_process,
        tasks=tasks,
        max_workers=args.workers,
        verbose=args.verbose,
        sleep=args.interval
    ).forever_run()


if __name__ == '__main__':
    try:
        main()
    except Exception as e:
        pawn.console.log(e)




#!/usr/bin/env python3
import argparse
from pawnlib.builder.generator import generate_banner
from pawnlib.__version__ import __version__
from pawnlib.output.color_print import *
from pawnlib.config import pawnlib_config as pawn, pconf
try:
    from pawnlib.utils import icx_signer
    icx_signer_loaded = True
except ImportError:
    icx_signer_loaded = False
    pass
from pawnlib.utils.operate_handler import run_with_keyboard_interrupt
from pawnlib.input import PromptWithArgument

__description__ = "This command can read and write to the wallet."

if not icx_signer_loaded:
    pawn.console.log("[red]Required packages - coincurve, eth_keyfile ")


def get_parser():
    parser = argparse.ArgumentParser(description='ICON')
    parser = get_arguments(parser)
    return parser


def get_arguments(parser):
    parser.add_argument(
        'sub_command',
        help='create, load',
        nargs='?'
    )

    parser.add_argument('-pk', '--private-key', metavar='private_key', help='A private key string in hexadecimal. default: None',
                        default=None)
    parser.add_argument('-d', '--debug', action='store_true', help='debug mode. True/False', default=False)
    parser.add_argument('-k', '--keystore', metavar='keystore', help='keystore filename or keystore text')
    parser.add_argument('-p', '--password', metavar='password', help='keystore\'s password')
    parser.add_argument('-l', '--load-type', metavar='load_type', help='load type (file or text)', choices=["file", "text"])
    parser.add_argument('-ns', '--no-store', action='store_true', help='Do not save as a file.', default=False)
    parser.add_argument('--base-dir', metavar='base_dir', help='base directory', default=os.getcwd())

    return parser


def private_key_validator(private_key):
    if icx_signer.is_private_key(private_key) or private_key == "":
        return True
    return "Invalid private key "


def least_length_validator(text, length=1):
    if len(text) >= length or text != "":
        return True
    return f"Must be at least {length} word(s) "


def main():
    icx_signer.compressed = False
    banner = generate_banner(
        app_name="WALLET",
        author="jinwoo",
        description="Wallet tools",
        font="graffiti",
        version=f"pawnlib {__version__}"
    )
    print(banner)
    parser = get_parser()
    args, unknown = parser.parse_known_args()

    if args.sub_command == "wallet":
        args.sub_command = ""

    pawn.console.log(f"args = {args}")
    args.subparser_name = "wallet"
    pawn.set(
        PAWN_DEBUG=args.debug,
        data=dict(
            args=args
        )
    )
    pawn.console.log(args.sub_command)
    args.sub_command = PromptWithArgument(
        message="What do you want to do?",
        choices=
        [
            {"name": "Load wallet", "value": "load"},
            {"name": "Create wallet", "value": "create"},
        ],
        instruction="",
        max_height="40%",
        default="",
        argument="sub_command",
        verbose=0,
    ).fuzzy()
    pawn.console.log(f"sub_command = {args.sub_command}")
    wallet_cli = icx_signer.WalletCli(args=pconf().data.args)

    if args.sub_command == "load":
        wallet_cli.load()

    elif args.sub_command == "create":
        is_store_file = False  if args.no_store else True
        wallet_cli.create(is_store_file=is_store_file)


if __name__ == '__main__':
    main()



from .prompt import (
    PromptWithArgument,
    CompareValidator,
    NumberCompareValidator,
    StringCompareValidator,
    MinMaxValidator,
    HexValidator,
    PrivateKeyValidator,
    PrivateKeyOrJsonValidator,
    check_valid_private_length,
    get_operator_truth,
    prompt_with_keyboard_interrupt,
    inq_prompt,
    simple_input_prompt,
    json_input_prompt,
    change_select_pattern,
    fuzzy_prompt,
    is_args_namespace,
    is_data_args_namespace,
    parse_list,
    get_environment
)



#!/usr/bin/env python3
import argparse
from pawnlib.builder.generator import generate_banner
from pawnlib.__version__ import __version__ as _version
from pawnlib.config import pawn, pconf
from pawnlib.typing import StackList, list_to_oneline_string, str2bool
from pawnlib.resource import SystemMonitor, get_cpu_load, get_interface_ips, get_platform_info, get_mem_info, get_netstat_count
from pawnlib.output import is_file
import os
import re

from rich.live import Live
from rich.table import Table
from rich.align import Align
from rich.text import Text
from pawnlib.typing import todaydate
import time

__description__ = "This is a tool to measure your server's resources."


if str2bool(os.environ.get("IS_DOCKER")) and is_file("/rootfs/proc"):
    PROCFS_PATH = "/rootfs/proc"
    pawn.console.log("Running in Docker mode on " + PROCFS_PATH)
else:
    PROCFS_PATH = "/proc"


def get_parser():
    parser = argparse.ArgumentParser(description='monitor')
    parser = get_arguments(parser)
    return parser


def get_arguments(parser):
    parser.add_argument('command', help='command', type=str, nargs='?', default="")
    parser.add_argument('-c', '--config-file', type=str, help='config', default="config.ini")
    parser.add_argument('-v', '--verbose', action='count', help='verbose mode. view level (default: %(default)s)', default=1)
    parser.add_argument('-q', '--quiet', action='count', help='Quiet mode. Dont show any messages. (default: %(default)s)', default=0)
    parser.add_argument('-i', '--interval', type=float, help='interval sleep time seconds. (default: %(default)s)', default=1)
    parser.add_argument('-b', '--base-dir', type=str, help='base dir for httping (default: %(default)s)', default=os.getcwd())
    parser.add_argument('-t', '--print-type', type=str, help='printing type  %(default)s)', default="line", choices=["live", "tab", "line"])
    return parser


class CriticalText:
    def __init__(self, column="", value="",  cores=1, warning_percent=75, medium_percent=50, low_percent=30, align_space=0):
        self.column = column
        self.value = value
        self.number_value = self.extract_first_number(str(value))
        self.align_space = align_space

        self.critical_limit_dict = {
            "net_in": 100,
            "net_out": 100,
            "usr": 80,
            "sys": 80,
            "mem_used": 99.5,
            "disk_read":  100,
            "disk_write":  100,
            "load":  int(cores),
            "io_wait":  int(cores) * 2,
            "cached":  10,
        }
        self.warning_percent = warning_percent
        self.medium_percent = medium_percent
        self.low_percent = low_percent

    @staticmethod
    def extract_first_number(text):
        match = re.match(r'\d+(\.\d+)?', text)
        if match:
            return float(match.group())
        return None

    def check_limit(self):
        limit_value = self.critical_limit_dict.get(self.column)
        if limit_value:
            if self.number_value >= limit_value:
                return "bold red"
            elif self.number_value >= (limit_value * self.warning_percent / 100):
                return "#FF9C3F"
            elif self.number_value >= (limit_value * self.medium_percent / 100):
                return "yellow"
            elif self.number_value >= (limit_value * self.low_percent / 100):
                return "green"
        return "white"

    def return_text(self):
        return Text(f"{self.value:>{self.align_space}}", self.check_limit())
        # return self.format_value_color()

    def __str__(self):
        return Text(f"{self.value:>{self.align_space}}", self.check_limit())

    # @staticmethod
    def format_value_color(self):
        formatted_text = ""
        color = self.check_limit()
        parts = self.value.split()
        for part in parts:
            if any(char.isdigit() for char in part):
                formatted_text += f'[{color}]{part}[/{color}] '
            else:
                formatted_text += f'[gray]{part}[/gray]'
        return f"{formatted_text:>{self.align_space+(len(color)*2)+1}}"


def print_banner():
    banner = generate_banner(
        app_name=pconf().app_name,
        author="jinwoo",
        description=f"{__description__} \n"
                    f" - base_dir    : {pconf().args.base_dir} \n" 
                    f" - logs_dir    : {pconf().args.base_dir}/logs \n",
        font="graffiti",
        version=_version
    )
    print(banner)


def main():
    app_name = 'monitor'
    parser = get_parser()
    args, unknown = parser.parse_known_args()
    config_file = args.config_file

    is_hide_line_number = args.verbose > 2
    stdout = not args.quiet

    pawn.set(
        PAWN_CONFIG_FILE=config_file,
        PAWN_PATH=args.base_dir,
        PAWN_LOGGER=dict(
            log_level="INFO",
            stdout_level="INFO",
            log_path=f"{args.base_dir}/logs",
            stdout=stdout,
            use_hook_exception=True,
            show_path=False, #hide line numbers
        ),
        PAWN_CONSOLE=dict(
            redirect=True,
            record=True,
            log_path=is_hide_line_number, # hide line number on the right side
        ),
        app_name=app_name,
        args=args,
        try_pass=False,
        last_execute_point=0,
        data={
            "response_time": StackList(),
        },
        fail_count=0,
        total_count=0,

    )
    if args.verbose > 2:
        pawn.set(
            PAWN_LOGGER=dict(
                log_level="DEBUG",
                stdout_level="DEBUG",
            )
        )
    print_banner()
    lines = []
    system_info = get_platform_info()
    system_monitor = SystemMonitor(interval=args.interval, proc_path=PROCFS_PATH)
    table_title = f"Server status <{system_info.get('model')},  {system_info.get('cores')} cores, {get_mem_info().get('mem_total')} GB>"

    if args.print_type == "live":
        print_live_status(table_title=table_title,  system_info=system_info, system_monitor=system_monitor)
    # elif args.print_type == "tab":
    #     print_tabulate_status(system_monitor=system_monitor)
    elif args.print_type == "line":
        count = 0
        while True:
            columns, term_rows = os.get_terminal_size()

            data = get_resources_status(system_monitor=system_monitor, args=args)
            line = []
            columns = []
            for column_key, value in data.items():
                align_space = max([len(str(value)), len(column_key)]) + 2
                columns.append(f"[blue][u]{column_key:^{align_space}}[/u][/blue]")
                _value = CriticalText(column_key, value, cores=system_info.get('cores', 1), align_space=align_space).return_text()
                line.append(_value)
            # pawn.console.print(list_to_oneline_string(columns, " | ")) if count % 15 == 0 else None
            print_line_status(columns) if count % term_rows == 0 else None
            print_line_status(line)
            count += 1


def print_line_status(line):
    for cell in line:
        pawn.console.print(cell, end="│")
    print()


def print_live_status(table_title="",  system_info={}, system_monitor=None):
    lines = []

    with Live(console=pawn.console, refresh_per_second=1) as live_table:
        while True:
            columns, rows = os.get_terminal_size()
            diff_rows = len(lines) - rows
            table = Table(title=table_title)

            data = get_resources_status(system_monitor=system_monitor)
            line = []
            for column_key, value in data.items():
                table.add_column(column_key)
                line.append(CriticalText(column_key, value, cores=system_info.get('cores', 1)).return_text())
            lines.append(line)

            for line in lines:
                table.add_row(*line)

            if len(lines) >= rows - 5:
                lines.pop(0)
            if diff_rows > -6:
                del lines[:6]
            live_table.update(Align.center(table))


# def print_tabulate_status(system_monitor=None):
#     _data = []
#     while True:
#         data = get_resources_status(system_monitor=system_monitor)
#         headers = list(data.keys())
#         values = list(data.values())
#         _data.append(values)
#         print(tabulate(_data, headers=headers, tablefmt="grid"))
#
#     print()


def get_resources_status(system_monitor={}, args=None):

    if args.command == "net":
        netstat = get_netstat_count(proc_path=PROCFS_PATH)
        data = {
            "time": todaydate("time_sec"),
        }
        data.update(netstat.get('COUNT'))
        time.sleep(args.interval)
    else:
        memory = system_monitor.get_memory_status()
        network, cpu, disk = system_monitor.get_network_cpu_status()
        # memory_unit = memory.get('unit').replace('B', "")
        memory_unit = memory.get('unit')

        data = {
            "time": todaydate("time_sec"),
            "net_in": f"{network['Total'].get('recv'):.2f}M",
            "net_out": f"{network['Total'].get('sent'):.2f}M",
            "pk_in": f"{network['Total'].get('packets_recv')}",
            "pk_out": f"{network['Total'].get('packets_sent')}",
            "load": f"{get_cpu_load()['1min']}",
            "usr": f"{cpu.get('usr')}%",
            "sys": f"{cpu.get('sys')}%",
            "io_wait": f"{cpu.get('io_wait')}",
            "disk_read":  f"{disk['total'].get('read_mb')} M",
            "disk_write":  f"{disk['total'].get('write_mb')} M",
            "mem_total": f"{memory.get('total')} {memory_unit}",
            # "mem_used": f"{memory.get('percent')}%",
            "mem_free": f"{memory.get('free')} {memory_unit}",
            "cached": f"{memory.get('cached')} {memory_unit}",
        }
    return data


if __name__ == '__main__':
    try:
        main()
    except Exception as e:
        pawn.console.log(e)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import sys
import json
import glob
import yaml
from typing import Union, Any
from pawnlib.config.globalconfig import pawnlib_config as pawn
from pawnlib.output import color_print
from pawnlib.typing import converter
from rich.prompt import Confirm


def check_file_overwrite(filename, answer=None) -> bool:
    """
    Checks the existence of a file.

    :param filename:
    :param answer:
    :return:

    Example:

        .. code-block:: python

            # touch sdsd

            from pawnlib.output import file
            file.check_file_overwrite(filename="sdsd")
            # >>   File already exists => sdsd
            # >>  Overwrite already existing 'sdsd' file? (y/n)


    """
    if not filename:
        return False

    if is_file(filename):
        color_print.cprint(f"File already exists => {filename}", "green")
        if answer is None:
            answer = Confirm.ask(prompt=f"Overwrite already existing '{filename}' file?", default=False)

        if answer:
            color_print.cprint(f"Removing the existing file => {filename}", "green")
            os.remove(filename)
        else:
            color_print.cprint("Operation stopped.", "red")
            sys.exit(1)
    return True


def get_file_path(filename) -> dict:
    """

     This function is intended to return the file information

    :param filename:
    :return:

    Example:

        .. code-block:: python

            from pawnlib.output import file
            file.get_file_path("sample_file.txt")

            # >>
               {
                  dirname:
                  file: sample_file.txt
                  extension: txt
                  filename: sample_file.txt
                  full_path: /examples/sample_file.txt
               }



    """
    _dirname, _file = os.path.split(filename)
    _extension = os.path.splitext(filename)[1]
    _fullpath = get_abs_path(filename)
    return {
        "dirname": _dirname,
        "file": _file,
        "extension": _extension,
        "filename": filename,
        "full_path": _fullpath
    }


def get_file_extension(file_path):
    """
    Get the file extension from a file path.

    :param file_path: The path of the file.
    :return: The extension of the file.

    Example:

        .. code-block:: python

            utils.get_file_extension("/home/user/document.txt")
            # >> 'txt'

            utils.get_file_extension("/home/user/.hiddenfile")
            # >> ''

            utils.get_file_extension("/home/user/.hiddenfile.txt")
            # >> 'txt'
    """
    _, file_extension = os.path.splitext(file_path)
    return file_extension[1:]  # Remove the leading "."


def get_file_list(path="./", pattern="*", recursive=False):
    """
    Get the list of files in the specified directory that match the given pattern.

    :param path: The path of the directory to search. Default is current directory.
    :param pattern: The pattern to match. Default is "*", which matches all files.
    :param recursive: Whether to search subdirectories recursively. Default is False.

    :return: A list of file paths that match the pattern.

    Example:

        .. code-block:: python

            get_file_list(directory_path="./", pattern="*.py", recursive=True)
            # >> ['./main.py', './utils/helper.py']

            get_file_list(directory_path="./", pattern="*.txt", recursive=False)
            # >> ['./README.txt']

    """
    try:
        if recursive:
            search_pattern = os.path.join(path, '**', pattern)
        else:
            search_pattern = os.path.join(path, pattern)
        files = glob.glob(search_pattern, recursive=recursive)
        return files

    except OSError:
        print("Error reading directory '{}'.".format(path))
        return None


def get_parent_path(run_path=__file__) -> str:
    """
    Returns the parent path

    :param run_path: (str) Path of the file to get the parent path of.
    :return: (str) Parent path of the given file path.

    Example:

        .. code-block:: python

            get_parent_path(__file__)
            # >> '/path/to/parent/directory'
    """
    path = os.path.dirname(os.path.abspath(run_path))
    parent_path = os.path.abspath(os.path.join(path, ".."))
    return parent_path


def get_script_path(run_path=__file__):
    """
    Returns the script path.

    :param run_path: (str) Path of the file to get the script path of.
    :return: (str) Script path of the given file path.

    Example:

        .. code-block:: python

            get_script_path(__file__)
            # >> '/path/to/script/directory'
    """
    return os.path.dirname(run_path)


def get_real_path(run_path=__file__):
    """
    Returns the real path.

    :param run_path: (str) Path of the file to get the real path of.
    :return: (str) Real path of the given file path.

    Example:

        .. code-block:: python

            get_real_path(__file__)
            # >> '/path/to/real/directory'
    """
    path = os.path.dirname(get_abs_path(run_path))
    return path


def get_abs_path(filename) -> str:
    """
    Returns the absolute path.

    :param filename: (str) Name of the file to get the absolute path of.
    :return: (str) Absolute path of the given file name.

    Example:

        .. code-block:: python

            get_abs_path('example.txt')
            # >> '/path/to/example.txt'
    """
    return os.path.abspath(filename)


def is_binary_file(filename) -> bool:
    """
    Check if the file is binary.

    :param filename: (str) Name of the file to check.
    :return: (bool) True if the file is binary, False otherwise.

    Example:

        .. code-block:: python

            is_binary_file('example.txt')
            # >> False
    """
    text_chars = bytearray({7, 8, 9, 10, 12, 13, 27} | set(range(0x20, 0x100)) - {0x7f})
    fh = open(filename, 'rb').read(100)
    return bool(fh.translate(None, text_chars))


def is_file(filename: str) -> bool:
    """
    Check if the file exists.

    :param filename: (str) Name of the file to check.
    :return: (bool) True if the file exists, False otherwise.

    Example:

        .. code-block:: python

            is_file('example.txt')
            # >> True
    """

    if "*" in filename:
        if len(glob.glob(filename)) > 0:
            return True
        else:
            return False
    else:
        return os.path.exists(os.path.expanduser(filename))


def is_directory(path):
    """
    Check if the given path is a directory.

    :param path: The path to check.
    :return: True if the path is a directory, False otherwise.

    Example:

        .. code-block:: python

            check.is_directory("/home/user")
            # >> True

            check.is_directory("/home/user/myfile.txt")
            # >> False

    """
    return os.path.isdir(path)


def is_json(json_file: str) -> bool:
    """
    Validate the JSON.

    :param json_file: (str) Name of the JSON file to validate.
    :return: (bool) True if the JSON is valid, False otherwise.

    Example:

        .. code-block:: python

            is_json('example.json')
            # >> True
    """
    try:
        with open(json_file, 'r', encoding="utf-8-sig") as j:
            json.loads(j.read())
    except ValueError:
        return False
    return True


def open_json(filename: str):
    """
    Read the JSON file.

    :param filename: str, the name of the JSON file to be read.
    :return: dict, the contents of the JSON file as a dictionary.

    Example:

        .. code-block:: python

            json_data = open_json("data.json")
            # >> {'name': 'John', 'age': 30, 'city': 'New York'}
    """
    try:
        with open(filename, "r") as json_file:
            return json.loads(json_file.read())
    except Exception as e:
        pawn.error_logger.error(f"[ERROR] Can't open the json -> '{filename}' / {e}") if pawn.error_logger else False
        raise


def open_file(filename: str):
    """
    Read the file.

    :param filename: str, the name of the file to be read.
    :return: str, the contents of the file as a string.

    Example:

        .. code-block:: python

            file_contents = open_file("example.txt")
            # >> 'This is an example file.\nIt contains some text.\n'
    """
    try:
        with open(filename, "r") as file_handler:
            return file_handler.read()
    except Exception as e:
        pawn.error_logger.error(f"[ERROR] Can't open the file -> '{filename}' / {e}") if pawn.error_logger else False
        raise


def open_yaml_file(filename: str):
    """
    Read the YAML file.

    :param filename: str, the name of the YAML file to be read.
    :return: dict, the contents of the YAML file as a dictionary.

    Example:

        .. code-block:: python

            yaml_data = open_yaml_file("data.yaml")
            # >> {'name': 'John', 'age': 30, 'city': 'New York'}
    """
    read_yaml = open_file(filename)
    return yaml.load(read_yaml, Loader=yaml.FullLoader)


def write_file(filename: str, data: Any, option: str = 'w', permit: str = '664'):
    """
    Write data to a file.

    :param filename: The name of the file to write.
    :param data: The data to write to the file.
    :param option: The write mode. Default is 'w'.
    :param permit: The permission of the file. Default is '664'.

    Example:
        .. code-block:: python

            write_file('test.txt', 'Hello, World!')
            # >> 'Write file -> test.txt, 13'

    :return: A string indicating the success or failure of the write operation.
    """

    with open(filename, option) as outfile:
        outfile.write(data)
    os.chmod(filename, int(permit, base=8))
    if os.path.exists(filename):
        return "Write file -> %s, %s" % (filename, converter.get_size(filename))
    else:
        return "write_file() can not write to file"


def write_json(filename: str, data: Union[dict, list], option: str = 'w', permit: str = '664', force_write: bool = True, json_default=None):
    """
    Write JSON data to a file.

    :param filename: The name of the file to write.
    :param data: The JSON data to write to the file.
    :param option: The write mode. Default is 'w'.
    :param permit: The permission of the file. Default is '664'.
    :param force_write: Whether to force the write operation. Default is True.
    :param json_default: A function to convert non-serializable objects to a serializable format. Default is None.

    Example:
        .. code-block:: python

            data = {'name': 'John', 'age': 30}
            write_json('test.json', data)
            # >> 'Write json file -> test.json, 23'

    :return: A string indicating the success or failure of the write operation.
    """

    def _json_default(obj):
        if hasattr(obj, 'to_json'):
            return obj.to_json()
        else:
            return str(obj)

    if not force_write:
        _json_default = None

    if json_default:
        _json_default = json_default

    with open(filename, option) as outfile:
        json.dump(data, outfile, default=_json_default)

    os.chmod(filename, int(permit, base=8))
    if os.path.exists(filename):
        return "Write json file -> %s, %s" % (filename, converter.get_size(filename))
    else:
        return "write_json() can not write to json"


def represent_ordereddict(dumper, data):
    """
    Represents an OrderedDict for YAML.

    :param dumper: A YAML dumper instance.
    :param data: An OrderedDict instance.

    Example:

        .. code-block:: python

            import yaml
            from collections import OrderedDict

            data = OrderedDict()
            data['key1'] = 'value1'
            data['key2'] = 'value2'

            yaml.add_representer(OrderedDict, represent_ordereddict)
            print(yaml.dump(data))

            # >> "key1: value1\nkey2: value2\n"

    """
    value = []
    node = yaml.nodes.MappingNode("tag:yaml.org,2002:map", value)
    for key, val in data.items():
        node_key = dumper.represent_data(key)
        node_val = dumper.represent_data(val)
        value.append((node_key, node_val))
    return node


def write_yaml(filename: str, data: Union[dict, list], option: str = 'w', permit: str = '664'):
    """
    Write YAML data to a file.

    :param filename: The name of the file to write.
    :param data: The YAML data to write to the file.
    :param option: The write mode. Default is 'w'.
    :param permit: The permission of the file. Default is '664'.

    Example:
        .. code-block:: python

            data = {'name': 'John', 'age': 30}
            write_yaml('test.yaml', data)
            # >> 'Write json file -> test.yaml, 23'

    :return: A string indicating the success or failure of the write operation.
    """
    yaml.add_representer(dict, represent_ordereddict)

    with open(filename, option) as outfile:
        yaml.dump(data, outfile)
    os.chmod(filename, int(permit, base=8))
    if os.path.exists(filename):
        return "Write json file -> %s, %s" % (filename, converter.get_size(filename))
    else:
        return "write_json() can not write to json"





"""Generators which yield an id to include in a JSON-RPC request."""
import sys
import os
import itertools
from string import ascii_lowercase, digits, ascii_uppercase

from uuid import uuid4
from pawnlib.typing.converter import UpdateType, replace_ignore_char, flatten_dict
import math
import random
import json

from functools import reduce, partial
from typing import Any, Dict, Iterator, Tuple, Union, Callable, Type
import binascii
import re
from argparse import ArgumentTypeError


class Null(object):
    """
    A Null object class as part of the Null object design pattern.
    """

    def __init__(self, *args, **kwargs):
        """
        Do nothing.
        """
        pass

    def __call__(self, *args, **kwargs):
        """
        Do nothing.
        @return: This object instance.
        @rtype: Null
        """
        return self

    def __getattr__(self, name):
        """
        Do nothing.
        @return: This object instance.
        @rtype: Null
        """
        return self

    def __setattr__(self, name, value):
        """
        Do nothing.
        @return: This object instance.
        @rtype: Null
        """
        return self

    def __delattr__(self, name):
        """
        Do nothing.
        @return: This object instance.
        @rtype: Null
        """
        return self

    def __repr__(self):
        """
        Null object string representation is the empty string.
        @return: An empty string.
        @rtype: String
        """
        return ''

    def __str__(self):
        """
        Null object string representation is the empty string.
        @return: An empty string.
        @rtype: String
        """
        return ''

    def __bool__(self):
        """
        Null object evaluates to False.
        @return: False.
        @rtype: Boolean
        """
        return False


class Counter:
    def __init__(self,
                 start: Union[float, int] = 1,
                 stop: Union[float, int] = 10,
                 step: Union[float, int] = 1,
                 convert_func: Callable = int,
                 kwargs: Dict = None):
        """
        Count up

        :param start: Start Number ( float or int )
        :param stop: Stop Number ( float or int )
        :param step: Step Number ( float or int )
        :param convert_func: Functions to execute on increment
        :param kwargs: Arguments to the function to be executed on increment

        Example:

            .. code-block:: python

                from pawnlib.utils.generator import Counter, generate_hex

                for i in Counter(start=0, step=1, stop=4):
                    pawn.console.log(f"Counter => {i}")

                # Counter => 0
                # Counter => 1
                # Counter => 2
                # Counter => 3

                for i in Counter(start=0, step=1, stop=10, convert_func=generate_hex, kwargs={"zfill": 10}):
                    pawn.console.log(f"Hex Counter => {i}")

                # Hex Counter => 0000000000
                # Hex Counter => 0000000001
                # Hex Counter => 0000000002
                # Hex Counter => 0000000003

        """
        if kwargs is None:
            kwargs = {}

        self.start = start
        self.stop = stop
        self.step = step
        self.convert_func = convert_func
        self.kwargs = kwargs

    def __iter__(self):
        return self

    def __str__(self):
        element_count = math.ceil((self.stop - self.start) / self.step)
        if getattr(self.convert_func, '__name__', None):
            function_name = f"{self.convert_func.__name__}()"
        else:
            function_name = self.convert_func
        return f"<Counter> start={self.start}, step={self.step}, stop={self.stop}, " \
               f"element_count={element_count}, convert_func={function_name}"

    def __next__(self):
        if self.start < self.stop:
            r = self.start
            self.start += self.step
            if isinstance(self.convert_func, Callable):
                return self.convert_func(r, **self.kwargs)
            return r
        else:
            raise StopIteration


class GenMultiMetrics:

    def __init__(self, tags, measurement, is_flatten=True, is_debug=False, structure_types=None, ignore_fields=None, uid=None):
        self.tags = tags
        self.measurement = measurement
        if uid is None:
            self.uid = id_generator()
        else:
            self.uid = uid
        self.is_flatten = is_flatten
        self.is_debug = is_debug
        self.structure_types = structure_types
        self.update_type = UpdateType(structure_types=self.structure_types)
        self.ignore_fields = ignore_fields
        self.return_value = {}

    def _set_default_metric(self):
        default_metric = {
            "measurement": self.measurement,
            # "tags": self.tags,
            "tags": {},
            "fields": {},
        }
        default_metric['tags'].update(self.tags)
        return default_metric

    def push(self, metric_key, key, value, tags=None):
        uid_key = f'{metric_key}_{self.uid}'
        key = replace_ignore_char(key)
        value = replace_ignore_char(value, replace_str="")

        if not value:
            value = 0

        if self.return_value.get(uid_key) is None:
            self.return_value[uid_key] = self._set_default_metric()
            if tags:
                self.return_value[uid_key]["tags"].update(tags)

        if self.ignore_fields is not None and key in self.ignore_fields:
            pass
        elif value or value == 0:
            if self.is_flatten and isinstance(value, dict):
                value = flatten_dict({key: value}, "_")
                self.return_value[uid_key]["fields"].update(value)
            else:
                value = self.update_type.assign_kv(key, value)
                if self.is_debug:
                    value = f"{value} ({type(value)})"
                self.return_value[uid_key]["fields"][key] = value

    def get(self):
        return list(self.return_value.values())


def generate_number_list(start=10000, count=100, convert_func=int):
    """
    Generate a list of numbers based on the given parameters.

    :param start: (int) Starting number (default: 10000)
    :param count: (int) Number of numbers to generate (default: 100)
    :param convert_func: (function) Function to convert the numbers (default: int)
    :return: (list) List of generated numbers

    Example:

        .. code-block:: python

            # Generate a list of integers from 10000 to 10099
            generate_number_list()

            # Generate a list of floats from 10000.0 to 10099.0
            generate_number_list(start=10000.0, convert_func=float)

    """
    result = []
    end = start + count + 1
    for i in range(start, end):
        numbering = convert_func(i)
        result.append(numbering)
    return result


def id_generator(size=8, chars=ascii_uppercase + digits):
    """
    this function will be generated random id

    :param size:
    :param chars:
    :return:

    :Example

        .. code-block:: python

            # >> 00ZP5YRLRT1Y
    """
    return ''.join(random.choice(chars) for _ in range(size))


def uuid_generator(size: int = 8, count: int = 4, separator: str = "-"):
    """
    this function will be generated random uuid

    :param size:
    :param count:
    :param separator:
    :return:

    :Example

        .. code-block:: python

            # >> KFXSYSVJHPE6-83KZTPKY9NL3-ZHRFUV7QRWWJ-GRWVPB6C5SM8

    """
    return separator.join([id_generator(size) for _ in range(count)])


def decimal(start: int = 1) -> Iterator[int]:
    """
    Increments from `start`.
    e.g. 1, 2, 3, .. 9, 10, 11, etc.

    :param start: start: The first value to start with.
    :return:
    """

    return itertools.count(start)


def hexadecimal(start: int = 1) -> Iterator[str]:
    """
    Incremental hexadecimal numbers.
    e.g. 1, 2, 3, .. 9, a, b, etc.

    :param start: The first value to start with.
    """

    while True:
        yield "%x" % start
        start += 1


def uuid() -> Iterator[str]:
    """
    Unique uuid ids.

    Example:

        '9bfe2c93-717e-4a45-b91b-55422c5af4ff'
    """
    while True:
        yield str(uuid4())


class Sentinel:
    def __init__(self, name: str):
        self.name = name

    def __repr__(self) -> str:
        return f"<{sys.intern(str(self.name)).rsplit('.', 1)[-1]}>"


def compose(*fs: Callable[..., Any]) -> Callable[..., Any]:
    def compose2(f: Callable[..., Any], g: Callable[..., Any]) -> Callable[..., Any]:
        return lambda *a, **kw: f(g(*a, **kw))
    return reduce(compose2, fs)


def request_pure(
        id_generator_func: Iterator[Any],
        method: str,
        params: Union[Dict[str, Any], Tuple[Any, ...]],
        id: Any,
) -> Dict[str, Any]:
    return {
        "jsonrpc": "2.0",
        "method": method,
        **(
            {"params": list(params) if isinstance(params, tuple) else params}
            if params
            else {}
        ),
        "id": id if id != "<NO_ID>" else next(id_generator_func),
    }


def request_impure(
        id_generator_func: Iterator[Any],
        method: str,
        params: Union[Dict[str, Any], Tuple[Any, ...], None] = None,
        id: Any = "<NO_ID>",
) -> Dict[str, Any]:
    return request_pure(
        id_generator_func or decimal(), method, params or (), id
    )


def json_rpc(
        method: str = "",
        params: Union[Dict[str, Any], Tuple[Any, ...], None] = None,
        id: Any = "<NO_ID>",
        dumps: bool = False,
):
    """

    :param method:
    :param params:
    :param id:
    :param dumps: to json string
    :return:


    Example:

        .. code-block:: python

            from pawnlib.typing import generator

            generator.json_rpc(method="icx_sendTransaction", params={"data": "ddddd"})
            # >  {'jsonrpc': '2.0', 'method': 'icx_sendTranscation', 'params': {'data': 'ddddd'}, 'id': 0}

            generator.json_rpc(method="icx_sendTransaction", params={"data": "ddddd"})
            # >  {'jsonrpc': '2.0', 'method': 'icx_sendTranscation', 'params': {'data': 'ddddd'}, 'id': 1}


    """

    if id != "<NO_ID>":
        pass
    else:
        id = increase_number(),

    if isinstance(id, tuple):
        id = id[0]

    return_dict = {
        "jsonrpc": "2.0",
        "method": method,
        **(
            {"params": list(params) if isinstance(params, tuple) else params}
            if params
            else {}
        ),
        "id": id
    }

    if dumps:
        return json.dumps(return_dict)

    return return_dict


def increase_number(c=itertools.count()):
    return next(c)


def increase_hex(c=itertools.count(), prefix="", zfill=0, remove_prefix=True):
    """
    Returns increase hex value

    :param c: itertools.count()
    :param prefix:
    :param zfill: adds zeros (0) at the beginning of the string, until it reaches the specified length.
    :param remove_prefix: remove prefix '0x' string
    :return:

    Example:

        .. code-block:: python

            from pawnlib.typing import generator

            generator.increase_hex()
            >> '0'

            generator.increase_hex()
            >> '1'


    """
    return generate_hex(next(c), prefix, zfill, remove_prefix)


def generate_hex(number=0, prefix="", zfill=0, remove_prefix=True):
    if remove_prefix:
        return f"{prefix}{hex(number).removeprefix('0x').zfill(zfill)}"
    else:
        return f"{prefix}{hex(number).zfill(zfill)}"


def generate_token_address(number=0, prefix="hx", zfill=40, remove_prefix=True):
    return generate_hex(number, prefix=prefix, zfill=zfill, remove_prefix=remove_prefix)


def increase_token_address(c=itertools.count(), prefix="hx", zfill=40, remove_prefix=True):
    """

    Returns increase token address

    :param c: itertools.count()
    :param prefix: prefix address
    :param zfill: adds zeros (0) at the beginning of the string, until it reaches the specified length.
    :param remove_prefix: remove prefix '0x' string

    Example:

        .. code-block:: python

            from pawnlib.typing import generator

            generator.increase_address()
            >> 'hx0000000000000000000000000000000000000000'

            generator.increase_address()
            >> 'hx0000000000000000000000000000000000000001'


    """
    return increase_hex(c, prefix=prefix, zfill=zfill, remove_prefix=remove_prefix)


def random_token_address(prefix="hx", nbytes=20):
    """

    Return a random hx address for icon network

    :return:

    Example:

        .. code-block:: python

            from pawnlib.typing import generator

            generator.random_token_address()
            >>> 'hxa85cfbf976afba6fd5880c89372cf9f253c4a1c9'


    """
    return f"{prefix}{token_hex(nbytes)}"


def random_private_key(nbytes=32):
    """
    :return:
    """
    bytes_key = os.urandom(nbytes) # or b"-B\x99\x99...xedy" + os.urandom(18)
    return bytes_key.hex()


def token_bytes(nbytes):
    """
    Return a random byte string containing *nbytes* bytes.
    If *nbytes* is ``None`` or not supplied, a reasonable
    default is used.

    :param nbytes:
    :return:


     Example:

        .. code-block:: python

            from pawnlib.typing import generator

            generator.token_bytes(16)
            >>> b'\\xebr\\x17D*t\\xae\\xd4\\xe3S\\xb6\\xe2\\xebP1\\x8b'

    """
    return os.urandom(nbytes)


def token_hex(nbytes):
    """

    Return a random text string, in hexadecimal.
    The string has *nbytes* random bytes, each byte converted to two
    hex digits.  If *nbytes* is ``None`` or not supplied, a reasonable
    default is used.

    :param nbytes:
    :return:

    Example:

        .. code-block:: python

            from pawnlib.typing import generator
            generator.token_hex(16)
            >>> 'f9bf78b9a18ce6d46a0cd2b0b86df9da'

    """

    return binascii.hexlify(token_bytes(nbytes)).decode('ascii')


def parse_regex_number_list(string):
    """
    Parse the list of numbers with regex.

    :param string:
    :return:

    Example:

        .. code-block:: python

            from pawnlib.typing import generator
            generator.parse_regex_number_list("1-6")

            >>> [ 1, 2, 3, 4, 5, 6]



    """
    m = re.match(r'(\d+)(?:-(\d+))?$', string)
    if not m:
        raise ArgumentTypeError("'" + string + "' is not a range of number. Expected forms like '0-5' or '2'.")
    start = m.group(1)
    end = m.group(2) or start
    return list(range(int(start, 10), int(end, 10)+1))


request_natural = partial(request_impure, decimal())
generate_json_rpc = compose(json.dumps, request_natural)



import logging
import os
import configparser
import sys
from uuid import uuid4
from pathlib import Path
from typing import Optional, Callable
from collections import namedtuple
from collections.abc import Mapping
import pawnlib.config.configure
from pawnlib.__version__ import __title__, __version__
from pawnlib.config.__fix_import import Null
from pawnlib.config.console import Console
from collections import OrderedDict
from rich.traceback import install as rich_traceback_install
import copy
from types import SimpleNamespace
from functools import partial
from rich import inspect as rich_inspect
import collections
import re


class NestedNamespace(SimpleNamespace):
    @staticmethod
    def _______map_____entry(entry):
        if isinstance(entry, dict):
            return NestedNamespace(**entry)

        return entry

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        for key, val in kwargs.items():
            if type(val) == dict:
                setattr(self, key, NestedNamespace(**val))
            elif type(val) == list:
                setattr(self, key, list(map(self._______map_____entry, val)))

    def _keys(self) -> list:
        return list(self.__dict__.keys())

    def _values(self) -> list:
        return list(self.__dict__.values())

    def as_dict(self) -> dict:
        return self._namespace_to_dict(self.__dict__)

    def _namespace_to_dict(self, _dict):
        dict2 = {}
        for k in _dict.keys():
            if isinstance(_dict[k], dict) or isinstance(_dict[k], NestedNamespace):
                dict2[k] = self._namespace_to_dict(_dict[k]._asdict())
            else:
                dict2[k] = _dict[k]
        return dict2

    def _asdict(self) -> dict:
        return self.__dict__

    # def __repr__(self):
    #     'Return a nicely formatted representation string'
    #     field_names = tuple(self.as_dict().keys())
    #     field_values = tuple(self.as_dict().values())
    #     indent = "\n    "
    #     # indent = ""
    #     repr_fmt = '(' + ', '.join(f'{indent}{name}=%r' for name in field_names) + f'{indent})'
    #     return self.__class__.__name__ + repr_fmt % field_values

    def __repr__(self, indent=4):
        result = self.__class__.__name__ + '('
        items_len = len(self.__dict__)
        _first = 0
        _indent_space = ''

        for k, v in self.__dict__.items():
            if _first == 0 and items_len > 0:
                result += "\n"
                _first = 1
            if k.startswith('__'):
                continue
            if isinstance(v, NestedNamespace):
                value_str = v.__repr__(indent + 4)
            else:
                value_str = str(v)

            if k and value_str:
                if _first:
                    _indent_space = ' ' * indent
                result += _indent_space + k + '=' + value_str + ",\n"
        result += ' ' * (len(_indent_space) - 4) + f')'
        return result


def nestednamedtuple(dict_items: dict, ignore_keys: list = []) -> namedtuple:
    """
    Converts dictionary to a nested namedtuple recursively.


    :param: dictionary: Dictionary to convert into a nested namedtuple.

    :example:

        .. code-block:: python

            from pawnlib.config.globalconfig import nestednamedtuple
            nt = nestednamedtuple({"hello": {"ola": "mundo"}})
            print(nt) # >>> namedtupled(hello=namedtupled(ola='mundo'))

    """
    dictionary = copy.copy(dict_items)
    if isinstance(dictionary, Mapping) and not isinstance(dictionary, fdict):
        # for ignore_type in ["configparser.SectionProxy", "configparser.ConfigParser"]:
        for ignore_type in ["configparser.ConfigParser"]:
            if ignore_type in str(type(dictionary)):
                for key, value in list(dictionary.items()):
                    dictionary[key] = value
                return dictionary

        for key, value in list(dictionary.items()):
            # if  key in ignore_keys:
            #     dictionary[key] = value
            # else:
            dictionary[key] = nestednamedtuple(value)

        return namedtuple("namedtupled", dictionary)(**dictionary)
    elif isinstance(dictionary, list):
        return [nestednamedtuple(item) for item in dictionary]

    return dictionary


class fdict(dict):
    """

    Forced dictionary. Prevents dictionary from becoming a nested namedtuple.

    :example:

        .. code-block:: python

            from pawnlib.config.globalconfig import nestednamedtuple, fdict
            d = {"hello": "world"}
            nt = nestednamedtuple({"forced": fdict(d), "notforced": d})
            print(nt.notforced)    # >>> namedtupled(hello='world')
            print(nt.forced)       # >>> {'hello': 'world'}
    """
    pass


def singleton(class_):
    instances = {}

    def getinstance(*args, **kwargs):
        if class_ not in instances:
            instances[class_] = class_(*args, **kwargs)
        return instances[class_]

    return getinstance


class ConfigSectionMap(configparser.ConfigParser):
    """
    override configparser.ConfigParser

    Example:

        .. code-block:: python

            config = ConfigSectionMap()
            config.read('config.ini')

            config_file = config.as_dict()

    """

    def __init__(self):
        # https://stackoverflow.com/questions/47640354/reading-special-characters-text-from-ini-file-in-python
        super(configparser.ConfigParser, self).__init__(interpolation=None)

    def as_dict(self, section=None):
        d = dict(self._sections)
        if self._defaults:
            d["DEFAULT"] = self._defaults
        for k in d:
            d[k] = dict(self._defaults, **d[k])
            d[k].pop('__name__', None)
        if section:
            return d.get(section)
        return d

    def get_default(self):
        return dict(self._defaults)


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


# @singleton
class PawnlibConfig(metaclass=Singleton):
    def __init__(self, global_name="pawnlib_global_config", app_logger=Null(), error_logger=Null(), timeout=6000, debug=False):
        """
        This class can share variables using globals().

        :param global_name: Global Variable Name
        :param app_logger: global app logger
        :param error_logger: global error logger
        :param timeout: global timeout

        :example

                .. code :: python

                    # auto attach
                    from pawnlib.config.globalconfig import pawnlib_config as pwn
                    from pawnlib.output.file import get_real_path

                    pwn.set(
                        PAWN_LOGGER=dict(
                            app_name="default_app",
                            log_path=f"{get_script_path(__file__)}/logs",
                            stdout=True,
                            use_hook_exception=False,
                        ),
                        PAWN_DEBUG=True,
                        app_name=APP_NAME,
                        data={} # Global NameSpace
                    )


                .. code :: python

                    # attach logger
                    from pawnlib.config.globalconfig import pawnlib_config as pwn
                    from pawnlib.output.file import get_real_path

                    app_logger, error_logger = log.AppLogger(
                        app_name="default_app",
                        log_path=f"{get_real_path(__file__)}/logs",
                        stdout=True,
                        use_hook_exception=False,
                    ).get_logger()

                    pwn.set(
                        PAWN_APP_LOGGER=app_logger,
                        PAWN_ERROR_LOGGER=error_logger,
                        PAWN_DEBUG=True,
                        app_name=APP_NAME,
                        data={} # Global NameSpace
                    )

        """
        self.global_name = f"{global_name}_{uuid4()}"
        self.app_logger = app_logger
        self.error_logger = error_logger

        self.app_name = ""
        self.timeout = timeout
        self.verbose = 0
        self.debug = debug

        self.version = f"{__title__}/{__version__}"
        self.env_prefix = "PAWN"
        self._environments = {}
        self.data = NestedNamespace()

        self._current_path: Optional[Path] = None

        self._config_file = None
        self._pawn_configs = {}
        self.console = Null()
        self.console_options = None
        self.stdout_log_formatter = None
        self._none_string = "____NONE____"

        self._loaded = {
            "console": False,
            "on_ready": False

        }

        self._do_not_execute_namespace_keys = [f"{self.env_prefix}_LOGGER", f"{self.env_prefix}_CONSOLE"]
        self.log_time_format = None

        globals()[self.global_name] = {}
        self._init_console(force_init=True)

    @staticmethod
    def inspect(*args, **kwargs):
        """
        Inspect function which can produce a report on any Python object, such as class, instance, or builtin.
        :param args:
        :param kwargs:
        :return:
        """
        return rich_inspect(*args, **kwargs)

    def _log_formatter(self, dt):

        if self.log_time_format.endswith('.%f'):
            return dt.strftime(self.log_time_format)[:-3]
        else:
            return dt.strftime(self.log_time_format)

    def _init_console(self, force_init=True):

        _console_options = dict(
            pawn_debug=self.debug,
            redirect=self.debug,  # <-- not supported by rich.console.Console
            record=True,
            soft_wrap=True,
            force_terminal=True,
            # log_time_format="[%Y-%m-%d %H:%M:%S.%f]"
            log_time_format=lambda dt: f"[{dt.strftime('%H:%M:%S,%f')[:-3]}]"
        )
        # print(self.console_options)
        if not self._loaded.get('console'):
            # There are visible problems with InquirerPy.
            _console_options['redirect'] = False

        if self._loaded.get('console') or force_init:
            if self.console_options:
                if self.console_options.get('log_time_format') and not isinstance(self.console_options.get('log_time_format'), Callable):
                    _log_time_format = self.console_options['log_time_format']
                    self.log_time_format = self.console_options['log_time_format']

                    if ".%f" in _log_time_format:
                        self.console_options['log_time_format'] = lambda dt: f"[{dt.strftime(_log_time_format)[:-3]}]"
                    else:
                        self.console_options['log_time_format'] = lambda dt: f"[{dt.strftime(_log_time_format)}]"
                    self.stdout_log_formatter = self.console_options['log_time_format']
                _console_options.update(self.console_options)
            # self.console_options = copy.deepcopy(_console_options)
            self.console = Console(**_console_options)

    def _load_config_file(self, config_path=None):
        if self._loaded['on_ready']:
            config = ConfigSectionMap()
            config.optionxform = str
            _config_filename = self.get_path(self._config_file)
            try:
                if _config_filename.is_file():
                    config.read(_config_filename)
                    self.set(PAWN_CONFIG=config.as_dict())
                else:
                    self.set(PAWN_CONFIG={})
                    self.console.debug(f"[bold red] cannot found config_file - {_config_filename}")
                for config_category, config_value in config.items():
                    lower_keys = [key.lower() for key in config[config_category].keys()]
                    duplicate_keys = _list_duplicates(lower_keys)
                    for conf_key, conf_value in config[config_category].items():
                        if conf_key.lower() in duplicate_keys:
                            self.console.log(f"[yellow]\[WARN] Similar keys exist in config.ini - \[{config_category}] {conf_key}={conf_value}")
            except Exception as e:
                self.console.log(f"[bold red]Error occurred while loading config.ini - {e}")

    def get_path(self, path: str = "") -> Path:
        """Get Path from the directory where the configure.json file is.

        :param path: file_name or path

        :return:

        """
        if self._current_path:
            root_path = Path(self._current_path)
        else:
            root_path = Path(os.path.join(os.getcwd()))
        return root_path.joinpath(path)

    @staticmethod
    def pawnlib_path():
        _dir = os.path.dirname(__file__)
        if "/config" in _dir:
            return _dir.replace("/config", "")
        return _dir

    @staticmethod
    def get_python_version():
        major, minor, micro = sys.version_info[:3]
        return f"Python {major}.{minor}.{micro} {sys.platform}"

    def init_with_env(self, **kwargs):
        """
        Initialize with environmental variables.

        :param kwargs: dict
        :return:
        """
        self.fill_config_from_environment()
        self.set(**kwargs)
        # self._load_config_file()
        self._loaded['on_ready'] = True
        self.console.debug(f"🐍 {self.get_python_version()}, ♙ {__title__.title()}/{__version__}, PATH={self.pawnlib_path()}")
        self._load_config_file()
        return self

    @staticmethod
    def str2bool(v):
        """
        This function is intended to return a boolean value.

        :param v:
        :return:
        """
        true_list = ("yes", "true", "t", "1", "True", "TRUE")
        if type(v) == bool:
            return v
        if type(v) == str:
            return v.lower() in true_list
        return eval(f"{v}") in true_list

    def fill_config_from_environment(self):
        """
        Initialize with environmental variables.

        .. code :: python

            # default environments

            PAWN_INI = False
            PAWN_DEBUG = False
            PAWN_VERBOSE = 0
            PAWN_TIMEOUT = 7000
            PAWN_APP_LOGGER = ""
            PAWN_ERROR_LOGGER = ""
            PAWN_VERSION =
            PAWN_GLOBAL_NAME = pawnlib_global_config_UUID

        :return:
        """
        default_structure = {
            "VERBOSE": {
                "type": int,
                "default": 0,
            },
            "INI": {
                "type": self.str2bool,
                "default": False,
            },
            "CONFIG_FILE": {
                "type": str,
                "default": "config.ini",
            },
            "DEBUG": {
                "type": self.str2bool,
                "default": False,
            },
            "TIMEOUT": {
                "type": int,
                "default": 7000,
            },
            "APP_LOGGER": {
                "type": str,
                "default": ""
            },
            "ERROR_LOGGER": {
                "type": str,
                "default": ""
            },
            "LOGGER": {
                "type": dict,
                "default": {}
            },
            "VERSION": {
                "type": str,
                "default": self.version.title()
            },
            "GLOBAL_NAME": {
                "type": str,
                "default": self.global_name
            },
            "CONSOLE": {
                "type": dict,
                "default": {}
            },
            "LINE": {
              "type": self.str2bool,
              "default": True,
            },
            "PATH": {
                "type": str,
                "default": Path(os.path.join(os.getcwd()))
            },
            "TIME_FORMAT": {
                "type": str,
                "default": "%H:%M:%S.%f"
            },
        }
        mandatory_environments = list(default_structure.keys())

        for environment in mandatory_environments:
            environment_name = f"{self.env_prefix}_{environment}"
            environment_value = os.getenv(environment_name)
            filled_environment_value = ""
            if default_structure.get(environment):
                required_type = default_structure[environment].get("type", None)
                if required_type is None:
                    self.console.log(f"[red]{environment_name} type is None. Required type")
                if environment_value in [None, 0, ""]:
                    filled_environment_value = default_structure[environment].get("default")
                elif required_type:
                    filled_environment_value = required_type(environment_value)
            self._environments[environment_name] = {
                "input": os.getenv(environment_name),
                "value": filled_environment_value,
            }
            self.set(**{environment_name: filled_environment_value})
            if isinstance(self.verbose, int) and self.verbose >= 3:
                self.console.debug(f"{environment_name}={filled_environment_value} (env={environment_value})")

    def make_config(self, dictionary: Optional[dict] = None, **kwargs) -> None:
        """Creates a global configuration that can be accessed anywhere during runtime.
        This function is a useful replacement to passing configuration classes between classes.
        Instead of creating a `Config` object, one may use :func:`make_config` to create a
        global runtime configuration that can be accessed by any module, function, or object.

        :param dictionary: Dictionary to create global configuration with.
        :param kwargs: Arguments to make global configuration with.

        Example:

            .. code-block:: python

                from pawnlib.config.globalconfig import PawnlibConfig
                PawnlibConfig.make_config(hello="world")


        """
        dictionary = dictionary or {}
        globals()[self.global_name] = {**dictionary, **kwargs}

    def get(self, key=None, default=None):
        """
        This method is intended to return a key value

        :param key:
        :param default:
        :return:

        Example:

            .. code-block:: python

                from pawnlib.config.globalconfig import pawnlib_config
                pawnlib_config.set(hello="world")
                pawnlib_config.get("hello")

        """
        if self.global_name in globals():
            return globals()[self.global_name].get(key, default)
        return default

    def set(self, **kwargs):
        """
        This method is intended to store key values.

        :param kwargs:
        :return:

        Example:

            .. code-block:: python

                from pawnlib.config.globalconfig import pawnlib_config
                pawnlib_config.set(hello="world")
                pawnlib_config.get("hello")

        """
        priority_keys = [f"{self.env_prefix}_PATH", f"{self.env_prefix}_TIME_FORMAT", f"{self.env_prefix}_DEBUG", f"{self.env_prefix}_VERBOSE"]
        order_dict = OrderedDict(kwargs)

        def _enforce_set_value(source_key=None, target_key=None, target_dict=None):
            if kwargs.get(source_key):
                if isinstance(target_dict, dict) and not target_dict.get(target_key):
                    target_dict[target_key] = kwargs[source_key]
                    if isinstance(self.verbose, int) and self.verbose >= 3:
                        self.console.debug(f'set => {target_key}={kwargs[source_key]}')

        for priority_key in priority_keys:
            if order_dict.get(priority_key):
                order_dict.move_to_end(key=priority_key, last=False)

        if self.global_name in globals():
            for p_key, p_value in order_dict.items():
                if self._environments.get(p_key, self._none_string) != self._none_string \
                        and self._environments[p_key].get("input"):
                    if self._environments[p_key].get('input') and \
                            self._environments[p_key].get('value') != p_value:
                        self.console.log(f"[yellow][WARN] Environment variables and settings are different. "
                                         f"'{p_key}': {self._environments[p_key]['value']}(ENV) != {p_value}(Config)")
                        p_value = self._environments[p_key]['value']

                if p_key == f"{self.env_prefix}_LOGGER" and p_value:
                    if isinstance(p_value, dict):
                        from pawnlib.utils.log import AppLogger
                        if p_value.get('app_name') is None and kwargs.get('app_name'):
                            p_value['app_name'] = kwargs['app_name']
                            self.app_name = kwargs['app_name']
                        _enforce_set_value(source_key=f'{self.env_prefix}_TIME_FORMAT', target_key='stdout_log_formatter', target_dict=p_value)
                        self.app_logger, self.error_logger = AppLogger(**p_value).get_logger()

                elif p_key == f"{self.env_prefix}_APP_LOGGER" and p_value:
                    self.app_logger = self._check_logger_not_null(p_key, p_value)
                elif p_key == f"{self.env_prefix}_ERROR_LOGGER" and p_value:
                    self.error_logger = self._check_logger_not_null(p_key, p_value)
                elif p_key == f"{self.env_prefix}_DEBUG":
                    self.debug = self.str2bool(p_value)
                    self.console.pawn_debug = self.str2bool(p_value)
                    if self.debug:
                        rich_traceback_install(show_locals=True, width=160)
                        if self.app_logger:
                            set_debug_logger(self.app_logger)
                elif p_key == f"{self.env_prefix}_LINE":
                    _hide_log_path = {"log_path": p_value}
                    if isinstance(self.console_options, dict):
                        self.console_options.update(**_hide_log_path)
                    else:
                        self.console_options = _hide_log_path
                    self._init_console()
                    # self._loaded['console'] = True
                elif p_key == f"{self.env_prefix}_CONSOLE":
                    _enforce_set_value(source_key=f'{self.env_prefix}_TIME_FORMAT', target_key='log_time_format', target_dict=p_value)
                    self.console_options = p_value
                    self._init_console()
                    self._loaded['console'] = True
                elif p_key == f"{self.env_prefix}_TIMEOUT":
                    self.timeout = p_value
                elif p_key == f"{self.env_prefix}_VERBOSE":
                    self.verbose = p_value
                elif p_key == f"{self.env_prefix}_PATH":
                    self._current_path = p_value
                    self._load_config_file()
                elif p_key == f"{self.env_prefix}_CONFIG_FILE":
                    self._config_file = p_value
                    self._load_config_file()

                elif p_key == "data" and p_value != self._none_string:
                    if isinstance(p_value, dict):
                        self.data = NestedNamespace(**p_value)
                    else:
                        self.console.log("[bold red] The data namespace value must be a Dict")
                    # else:
                    #     self.data = NestedNamespace()
                    #     setattr(self, p_key, p_value)
                    #     self.console.log(f"fff => {self.data}")
                    p_value = self.data

                globals()[self.global_name][p_key] = p_value

    def _check_logger_not_null(self, key, value):
        if type(value).__name__ == "Logger":
            return value
        else:
            self.console.debug(f"[red]Invalid Logger [/red] => {key}: {value} ({type(value)})")
            return Null()

    def increase(self, **kwargs):
        """
        Find the key and increment the number.

        :param kwargs:
        :return:

        Example:

            .. code-block:: python

                from pawnlib.config.globalconfig import pawnlib_config

                pawnlib_config.increase(count=1)
                print(pawnlib_config.get("count"))
                # >> 1

                pawnlib_config.increase(count=1)
                print(pawnlib_config.get("count"))

                # >> 2

                pawnlib_config.increase(count=10)
                print(pawnlib_config.get("count"))

                # >> 12


        """
        return self._modify_value(_command="increase", **kwargs) or 0

    def decrease(self, **kwargs):
        """
        Find the key and decrement the number.
        :param kwargs:
        :return:

        Example:

            .. code-block:: python

                from pawnlib.config.globalconfig import pawnlib_config

                pawnlib_config.set(count=100)
                print(pawnlib_config.get("count"))
                # >> 100

                pawnlib_config.decrease(count=1)
                print(pawnlib_config.get("count"))

                # >> 99

                pawnlib_config.decrease(count=10)
                print(pawnlib_config.get("count"))

                # >> 89

        """
        return self._modify_value(_command="decrease", **kwargs) or 0

    def append_list(self, **kwargs):
        """
        Find the key and append the value to list.
        :param kwargs:
        :return:

        Example:

            .. code-block:: python

                from pawnlib.config.globalconfig import pawnlib_config

                pawnlib_config.append_list(results="result1")
                pawnlib_config.append_list(results="result2")

                print(pawnlib_config.get("results"))

                # >> ['result1', 'result2']


        """

        return self._modify_value(_command="append_list", **kwargs) or []

    def remove_list(self, **kwargs):
        """
        Find the key and remove the value to list.
        :param kwargs:
        :return:

        Example:

            .. code-block:: python

                from pawnlib.config.globalconfig import pawnlib_config

                pawnlib_config.set(results=['result1', 'result2'])
                pawnlib_config.remove_list(results="result2")

                print(pawnlib_config.get("results"))

                # >> ['result1']

        """
        return self._modify_value(_command="remove_list", **kwargs) or []

    @staticmethod
    def _modify_value_initialize(_command=None):
        init_values = {
            "increase": 0,
            "decrease": 0,
            "append_list": [],
            "remove_list": [],
        }
        return init_values.get(_command)

    def _modify_value(self, _command=None, **kwargs):
        """
        Find the key and modify the value.
        :param _command:
        :param kwargs:
        :return:
        """
        is_modify = False
        init_value = self._modify_value_initialize(_command=_command)

        for key, value in kwargs.items():
            tmp_result = self.get(key=key, default="___NONE_VALUE___")

            if tmp_result == "___NONE_VALUE___":
                tmp_result = init_value

            if _command == "increase":
                if isinstance(tmp_result, int) or isinstance(tmp_result, float):
                    tmp_result += value
                    is_modify = True
            elif _command == "decrease":
                if isinstance(tmp_result, int) or isinstance(tmp_result, float):
                    tmp_result -= value
                    is_modify = True
            elif _command == "append_list" and isinstance(tmp_result, list):
                tmp_result.append(value)
                is_modify = True
            elif _command == "remove_list" and isinstance(tmp_result, list):
                tmp_result.remove(value)
                is_modify = True
            if is_modify:
                globals()[self.global_name][key] = tmp_result
                return tmp_result
        return init_value

    def __str__(self):
        return f"<{self.version.title()}>[{self.global_name}]\n{self.to_dict()}"

    def conf(self) -> namedtuple:
        """Access global configuration as a :class:`pawnlib.config.globalconfig.PawnlibConfig`.

        Example:

            .. code-block:: python

                from pawnlib.config.globalconfig import pawnlib_config
                print(pawnlib_config.conf().hello) # >>> 'world'

        """
        g = globals()
        if self.global_name in g:
            # return nestednamedtuple(g[self.global_name], ignore_keys=self._do_not_execute_namespace_keys)
            return NestedNamespace(**g[self.global_name])
        else:
            return NestedNamespace({})

    def to_dict(self) -> dict:
        """Access global configuration as a dict.

        Example:

            .. code-block:: python

                from pawnlib.config.globalconfig import pawnlib_config
                print(pawnlib_config.to_dict().get("hello")) # >>> 'world'

        """
        g = globals()
        if self.global_name in g:
            return g[self.global_name]
        else:
            return {}

    # def ns(self):
    #     return self.data
    #
    # def data_to_namespace(self):
    #
    #     g = globals()
    #     if self.global_name in g:
    #         return self.data
    #
    #     return self.data


def _list_duplicates(seq):
    seen = set()
    seen_add = seen.add
    # adds all elements it doesn't know yet to seen and all other to seen_twice
    seen_twice = set(x for x in seq if x in seen or seen_add(x))
    # turn the set into a list (as requested)
    return list(seen_twice)


def set_debug_logger(logger_name=None, propagate=0, get_logger_name='PAWNS', level='DEBUG'):
    if logger_name:
        __logger = logging.getLogger(get_logger_name)
        __logger.propagate = propagate
        __logger.setLevel(level)
        __logger.addHandler(logger_name)


pawnlib_config = PawnlibConfig(global_name="pawnlib_global_config").init_with_env()
pawn = pawnlib_config
pconf = partial(pawn.conf)
global_verbose = pawnlib_config.get('PAWN_VERBOSE', 0)



import re
import json
import sys
import socket
import ssl
from functools import partial
from datetime import datetime
from pawnlib.config.globalconfig import pawnlib_config as pawn, global_verbose, pconf, SimpleNamespace, Null
from pawnlib.output import (
    NoTraceBackException,
    dump, syntax_highlight, kvPrint, debug_logging,
    PrintRichTable, get_debug_here_info,
    print_json)
from pawnlib.resource import net
from pawnlib.typing import date_utils
from pawnlib.typing.converter import append_suffix, append_prefix, hex_to_number, FlatDict, FlatterDict, flatten, const, shorten_text
from pawnlib.typing.constants import const
from pawnlib.typing.generator import json_rpc, random_token_address, generate_json_rpc
from pawnlib.typing.check import keys_exists, is_int, is_float, list_depth, is_valid_token_address, sys_exit, is_hex, is_valid_tx_hash
from pawnlib.utils.operate_handler import WaitStateLoop
from pawnlib.output import pretty_json, align_text, get_file_extension, is_directory, is_file
from pawnlib.utils.in_memory_zip import gen_deploy_data_content
from websocket import create_connection, WebSocket, enableTrace
try:
    from pawnlib.utils import icx_signer
except ImportError:
    pass

from typing import Any, Dict, Iterator, Tuple, Union, Callable, Type
from collections import OrderedDict

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from enum import Enum, auto
import copy
import operator as _operator
import time
from dataclasses import dataclass, InitVar, field
import requests
from rich.prompt import Prompt, Confirm


ALLOWS_HTTP_METHOD = ["get", "post", "patch", "delete", "head", "put", "connect", "options", "trace", "patch"]
ALLOW_OPERATOR = ["!=", "==", ">=", "<=", ">", "<", "include", "exclude"]


class _ResponseWithElapsed(requests.models.Response):
    success = False
    error = None

    def __init__(self):
        super().__init__()

    def __repr__(self):
        return f"<Response [{self.status_code}], {self.elapsed}ms, succ={self.success}>"

    def __str__(self):
        return f"<Response [{self.status_code}], {self.elapsed}ms, succ={self.success}>"

    def as_dict(self):

        if not self.__dict__.get('result'):
            try:
                self.__dict__['result'] = self.__dict__['_content'].decode('utf-8')
            except:
                pass

        if self.__dict__.get('result') and isinstance(self.__dict__.get('result'), dict):
            self.__dict__['json'] = self.__dict__['result']
            self.__dict__['text'] = json.dumps(self.__dict__['result'])

        else:
            try:
                self.__dict__['json'] = json.loads(self.__dict__['result'])
            except:
                self.__dict__['json'] = {}
        return self.__dict__


requests.models.Response.__str__ = _ResponseWithElapsed.__str__
requests.models.Response.__repr__ = _ResponseWithElapsed.__repr__
requests.models.Response.error = _ResponseWithElapsed.error
requests.models.Response.success = _ResponseWithElapsed.success
requests.models.Response.as_dict = _ResponseWithElapsed.as_dict


class HttpResponse:
    def __init__(self, status_code=999, response=None, error=None, elapsed=None, success=False):
        self.status_code = status_code
        self.response = response
        self.error = error
        self.elapsed = elapsed
        self.success = success
        self.text = None

        if getattr(response, "text", None):
            self.response = response.text

        if self.response and self.response.json:
            self.json = self.response.json
        else:
            self.json = {}

        if self.error:
            self.ok = False
        else:
            self.ok = True

    def __str__(self):
        return f"<HttpResponse> status_code={self.status_code}, response={self.response}, error={self.error}"

    def __repr__(self):
        return f"<HttpResponse> {self.status_code}, {self.response}, {self.error}"

    def as_dict(self):
        return self.__dict__


class StrEnum(str, Enum):
    @staticmethod
    def _generate_next_value_(name, start, count, last_values):
        return name

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name


class AllowsHttpMethod(StrEnum):
    get = auto()
    post = auto()
    patch = auto()
    delete = auto()


class AllowsKey(StrEnum):
    status_code = auto()
    headers = auto()
    raw = auto()
    url = auto()
    reason = auto()
    http_version = auto()
    r_headers = auto()
    result = auto()
    elapsed = auto()


@dataclass
class NetworkInfo:
    network_name: str = "mainnet"
    platform: str = "icon"
    force: bool = False
    network_api: str = ""
    planet_api: str = ""
    nid: str = ""
    network: str = ""
    endpoint: str = ""
    symbol: str = ""
    valid_network: bool = False

    def __post_init__(self):

        self._platform_info = {
            "icon": {
                "symbol": "ICX",
                "network_info": {
                    "mainnet": {
                        "network_api": "https://ctz.solidwallet.io",
                        "nid": "0x1"
                    },
                    "lisbon": {
                        "network_api": "https://lisbon.net.solidwallet.io",
                        "nid": "0x2"
                    },
                    "techteam": {
                        "network_api": "https://techteam.net.solidwallet.io",
                        "nid": "0xa"
                    },
                    "cdnet": {
                        "network_api": "http://20.20.1.122:9000",
                        "nid": "0x53"
                    },
                }
            },
            "havah": {
                "symbol": "HVH",
                "network_info": {
                    "mainnet": {
                        "planet_api": "https://planet.havah.io",
                        "network_api": "https://ctz.havah.io",
                        "nid": "0x100"  # 256
                    },
                    "vega": {
                        "planet_api": "https://planet.vega.havah.io",
                        "network_api": "https://ctz.vega.havah.io",
                        "nid": "0x101"  # 257
                    },
                    "deneb": {
                        "planet_api": "https://planet.dev.havah.io",
                        "network_api": "https://ctz.dev.havah.io",
                        "nid": "0x110"  # 272
                    },
                    "svm_havah": {
                        "network_api": "http://20.20.1.153:9900",
                        "nid": "0x8361",
                    },
                }
            }
        }
        # self.network_name = network_name
        # self.platform = platform
        self.network_info = {}
        self.static_values = ["nid", "network_api", "endpoint"]
        if not self.force:
            self._initialize()

    def is_set_static_values(self):
        # missing_static_values = [static_value for static_value in self.static_values if not getattr(self, static_value, None)]
        # if missing_static_values:
        #     pawn.console.debug(f"[red]Missing static values: {missing_static_values}")
        return any(getattr(self, static_value, None) for static_value in self.static_values)

    def _get_network_info(self, network_name="", platform=""):
        if network_name:
            self.network_name = network_name
        if platform:
            self.platform = platform
        self.network_name = self.network_name.lower()
        self.platform = self.platform.lower()

        if self.network_name == "veganet":
            self.network_name = "vega"

        elif self.network_name == "denebnet":
            self.network_name = "deneb"

        if not self._platform_info.get(self.platform):
            raise ValueError(f"Allowed platform - values {list(self._platform_info.keys())}")
        if not self.is_set_static_values():
            _network_info = self._platform_info[self.platform].get('network_info')
            if isinstance(_network_info, dict) and not _network_info.get(self.network_name):
                raise ValueError(f"Allowed network_name in '{self.platform}' - values {list(_network_info.keys())}")
            self.network_info = self._platform_info[self.platform]['network_info'].get(self.network_name)

        self.symbol = self._platform_info[self.platform].get('symbol')

        if self.network_info:
            self.valid_network = True
            self.network_info['symbol'] = self.symbol
            self.network_info['network_name'] = self.network_name
            self.network_info['platform'] = self.platform
        else:
            self.network_info = self._extract_network_info()

        if not self.network_info.get('endpoint') and self.network_info.get('network_api'):
            self.network_info['endpoint'] = append_suffix(self.network_info['network_api'], "/api/v3")

    def _initialize(self, network_name="", platform=""):
        self._get_network_info(network_name, platform)
        if self.network_info and self.valid_network:
            self._set_attributes()

    def _set_attributes(self):
        this_data = self.network_info

        if this_data.get("network", None) is None:
            setattr(self, "network", self.network_name)
        for key, value in this_data.items():
            object.__setattr__(self, key, this_data.get(key))

    def set_network(self, network_name=None, platform="icon"):
        if network_name and platform:
            self._initialize_static_values()
            self._initialize(network_name=network_name, platform=platform)

    def _initialize_static_values(self):
        for static_value in self.static_values:
            setattr(self, static_value, None)

    def list(self) -> list:
        return list(self.network_info.keys())

    def get_platform_list(self) -> list:
        return list(self._platform_info.keys())

    def get_platform_info(self) -> dict:
        return self._platform_info

    def update_platform_info(self, data={}) -> dict:
        self._platform_info.update(data)
        return self._platform_info

    def add_network(self, platform, network_name, network_api, nid):
        if platform in self._platform_info:
            self._platform_info[platform]["network_info"][network_name.lower()] = {
                "network_api": network_api,
                "nid": nid
            }

    def get_network_list(self, platform="") -> list:
        if platform:
            self.platform = platform

        if self._platform_info.get(self.platform):
            return list(self._platform_info[self.platform]['network_info'].keys())

    def tuple(self) -> tuple:
        return tuple(self.network_info.keys())

    def to_dict(self):
        if self.network_info:
            return self.network_info
        return self.network_info

    def _extract_network_info(self):
        _static_network_info = {
            key: value for key, value in copy.deepcopy(self.__dict__).items()
            if key and value and key not in ['_platform_info', 'static_values', 'network_info']
        }
        return _static_network_info

    def _parse_network_info_string(self):
        return self.network_info if self.network_info else self._extract_network_info()

    def __repr__(self):
        return f"<{self.platform.upper()} {self.__class__.__name__}> network_info={self._parse_network_info_string()}"

    def __str__(self):
        return f"<{self.platform.upper()} {self.__class__.__name__}> {self._parse_network_info_string()}"


class IconRpcTemplates:
    requires_sign_method = ['icx_sendTransaction', 'icx_sendTransaction(SCORE)', 'icx_call']
    templates = {
        "main_api": {
            # "icx_getTotalSupply": json_rpc("icx_getTotalSupply"),
            "icx_getTotalSupply": {},
            "icx_getLastBlock": {},
            "icx_getBalance": {"params": {"address": ""}},
            "icx_getTransactionResult": {"params": {"txHash": ""}},
            "icx_getTransactionByHash": {"params": {"txHash": ""}},
            "icx_getBlockByHeight": {"params": {"height": ""}},
            "icx_getBlockByHash": {"params": {"hash": ""}},
            "icx_getScoreApi": {"params": {"address": ""}},
            "icx_getScoreStatus": {"params": {"address": ""}},
            "debug_getTrace": {"params": {"txHash": ""}},
            "debug_estimateStep": {"params": {}},
            "icx_getNetworkInfo": {},
            "icx_call": {"params": ""},
            "icx_sendTransaction": {
                "params": {
                    "from": "",
                    "to": "",
                    "nid": "",
                    "stepLimit": "",
                    "value": "",
                    "version": "0x3",
                    "nonce": "0x23",
                }
            },
            "icx_sendTransaction(SCORE)": {"method": "icx_sendTransaction"}
        },
        "IISS": {
            "setStake": dict(
                method="icx_sendTransaction",
                params={
                    "method": "setStake",
                    "params": {
                        "value": ""
                    }
                }
            ),
        }
    }

    def __init__(self, category=None, method=None):
        self.return_rpc = {}
        self._category = category
        self._method = method
        self._params = {}
        self.get_rpc()

    def update_template(self, new_template):
        self.templates.update(**new_template)

    def get_category(self):
        return list(self.templates.keys())

    def get_methods(self, category=None):
        methods = []
        for _category in self.get_category():
            if _category == category:
                return self.templates.get(_category).keys()
            methods += self.templates.get(_category).keys()
        return methods

    def create_rpc(self, params={}, method=None):

        pass

    def is_required_sign(self):
        if self._method in self.requires_sign_method:
            return True
        return False

    def load_template(self):
        if self._category:
            _template = self.templates.get(self._category)
        else:
            _template = {}
            for item in self.templates.values():
                _template.update(item)
        return _template

    def get_rpc(self, category=None, method=None):
        if category:
            self._category = category
        if method:
            self._method = method

        _template = self.load_template()
        if self._method and _template:
            _arguments = _template.get(method, {})
            if not isinstance(_arguments, dict):
                raise ValueError(f"[Template Error] Syntax Error -> category={self._category}, method={self._method}")

            if not self._method:
                raise ValueError(f"[Template Error] Required method ->  category={self._category}, method={self._method}")

            if _arguments.get('method'):
                self.return_rpc = json_rpc(**_arguments)
            else:
                self._method = _arguments.get('method', self._method)
                self._params = _arguments.get('params', {})
                self.return_rpc = json_rpc(method=self._method, params=self._params)

            # pawn.console.log(f"-- return_rpc {self.return_rpc}")

            return self.return_rpc
        return {}

    def get_required_params(self):
        return self._params


class IconRpcHelper:
    def __init__(self, url="", wallet=None, network_info: NetworkInfo = None, raise_on_failure=True, debug=False, required_sign_methods=None):
        self.wallet = wallet
        self.governance_address = None
        self.request_payload = None
        self.response = None
        self.network_info = network_info
        self.raise_on_failure = raise_on_failure
        self.debug = debug
        if required_sign_methods and isinstance(required_sign_methods, list):
            self.required_sign_methods = required_sign_methods
        else:
            self.required_sign_methods = ['set', 'register', "unregister", "claim", "vote", "apply", "remove", "cancel", "acceptScore", "reject"]

        if not url and self.network_info:
            url = self.network_info.network_api

        self.url = append_suffix(url, "/api/v3")
        self.debug_url = append_suffix(url, "/api/v3d")
        self.signed_tx = {}
        self._parent_method = ""
        self._can_be_signed = None
        self.on_error = False
        self.initialize()
        self._use_global_reqeust_payload = False
        self.global_reqeust_payload = {}
        self.score_api = {}

        self.default = {
            "stepLimit": hex(2500000)
        }

    def initialize(self):
        # self._set_governance_address()
        if self.network_info:
            pawn.console.debug(self.network_info)
        else:
            pawn.console.debug("Not found network_info")

        self.initialize_wallet()

    def initialize_wallet(self):
        if self.wallet and not isinstance(self.wallet, dict):
            pawn.console.debug("Loading wallet from icx_signer.load_wallet_key()")
            self.wallet = icx_signer.load_wallet_key(self.wallet)

    def _set_governance_address(self, method=None):
        if self.network_info and not self.governance_address:
            if self.network_info.platform == "havah":
                if method and method.startswith("get"):
                    self.governance_address = const.CHAIN_SCORE_ADDRESS
            else:
                self.governance_address = const.GOVERNANCE_ADDRESS

    def _decorator_enforce_kwargs(func):
        def from_kwargs(self, *args, **kwargs):
            func_name = func.__name__
            pawn.console.debug(f"Start '{func_name}' function")
            ret = func(self, *args, **kwargs)
            return ret

        return from_kwargs

    @staticmethod
    def _convert_str_to_dict(payload):
        if isinstance(payload, str):
            try:
                payload_dict = json.loads(payload)
            except:
                return payload
            return payload_dict

        return payload

    def _convert_valid_url_format(self, url=None):
        if url:
            self.url = url
        if self.url:
            self.url = append_http(append_api_v3(self.url))
        else:
            raise ValueError('Invalid url: %s' % self.url)

    def _convert_valid_payload_format(self, payload=None, method=None, params=None):
        if payload:
            _request_payload = self._convert_str_to_dict(payload)
        else:
            _request_payload = json_rpc(method=method, params=params)

        return _request_payload

    def rpc_call(self,
                 url=None,
                 method=None,
                 timeout=5000,
                 params: dict = {},
                 payload: dict = {},
                 print_error=False,
                 reset_error=True,
                 raise_on_failure=False,
                 store_request_payload=True,
                 http_method: Literal["get", "post", "patch", "delete"] = 'post',
                 ) -> dict:
        if url:
            _url = url
        else:
            _url = self.url

        if raise_on_failure:
            _raise_on_failure = raise_on_failure
        else:
            _raise_on_failure = self.raise_on_failure

        _url = append_http(append_api_v3(_url))
        _request_payload = self._convert_valid_payload_format(payload=payload, method=method, params=params)

        if reset_error:
            self.on_error = False

        if store_request_payload:
            self.request_payload = copy.deepcopy(_request_payload)

        self.response = CallHttp(
            url=_url,
            method=http_method,
            timeout=timeout,
            payload=_request_payload,
            raise_on_failure=_raise_on_failure,
        ).run().response.as_dict()

        if self.response.get('status_code') != 200:
            self.on_error = True
            self.print_error_message(print_error)

        return self.response.get('json')

    def print_error_message(self, print_error=True):
        if print_error:
            if self.response.get('json') and self.response['json'].get('error'):
                pawn.console.log(f"[red][ERROR][/red] payload={self.request_payload}")
                pawn.console.log(f"[red][ERROR] status_code={self.response['status_code']}, error={self.response['json']['error']}")
            elif self.response.get('status_code') == 999:
                pawn.console.log(f"[red][ERROR][/red] {self.response.get('error')}")
                # self.exit_on_failure(f"[red][ERROR][/red] {self.response.get('error')}")
            else:
                pawn.console.log(f"[red][ERROR][/red] status_code={self.response.get('status_code')}, text={self.response.get('text')}")

    def print_response(self, hex_to_int=False):
        if self.response.get('status_code') != 200:
            style = "red"
        else:
            style = "rule.line"
        pawn.console.rule(f"<Response {self.response.get('status_code')}>", align='right', style=style, characters="═")
        if self.response.get('json'):
            dump(self.response.get('json'), hex_to_int=hex_to_int)
        else:
            print(syntax_highlight(self.response.get('text'), name='html'))

    def print_request(self):
        pawn.console.print("")
        pawn.console.rule(f"<Request> {self.url}", align='left')
        pawn.console.print("")
        print(syntax_highlight(self.request_payload, line_indent='   '))

    def make_params(self, method=None, params={}):
        json_rpc(
            method=method,
            params=params
        )

    def _is_signable_governance_method(self, method):
        # if self.network_info.platform == "havah" or self.network_info.platform == "icon" and method:
        if method:
            for required_sign_method in self.required_sign_methods:
                if method.startswith(required_sign_method):
                    pawn.console.debug(f"{method}, It will be signed with the following. required_sign_methods={self.required_sign_methods}")
                    return True
        return False

    def create_governance_payload(self, method, params):

        if self._can_be_signed is None:
            self._can_be_signed = self._is_signable_governance_method(method)

        if self._can_be_signed:
            parent_method = "icx_sendTransaction"
        else:
            parent_method = "icx_call"
        _request_payload = self._convert_valid_payload_format(
            method=parent_method,
            params={
                "to": self.governance_address,
                "dataType": "call",
                "data": {
                    "method": method,
                    "params": params
                }
            },
        )
        return _request_payload

    def governance_call(self, url=None, method=None, params={}, governance_address=None,
                        sign=None, store_request_payload=True, is_wait=True, value="0x0", step_limit=None):
        if governance_address:
            self.governance_address = governance_address
        else:
            self._set_governance_address(method=method)

        if sign is not None:
            self._can_be_signed = sign
        pawn.console.debug(f"Does the method require a signature? _can_be_signed={self._can_be_signed}")
        _request_payload = self.create_governance_payload(method, params)

        if self._can_be_signed:
            # if _request_payload['params'].get('value', None):
            _request_payload['params']['value'] = value
            self.sign_tx(payload=_request_payload, step_limit=step_limit)
            response = self.sign_send(is_wait=is_wait)
            return response
        else:
            response = self.rpc_call(
                url=url,
                payload=_request_payload,
                print_error=True,
                store_request_payload=store_request_payload,
            )
            return response.get('result', {})

    def create_deploy_payload(self, src="", params={}, governance_address=None):
        self.governance_address = governance_address if governance_address else const.CHAIN_SCORE_ADDRESS
        file_extension_mapping = {
            "jar": "java"
        }
        _file_extension = get_file_extension(src) if is_file(src) else None
        _file_type = file_extension_mapping.get(_file_extension, "python")

        if not _file_extension:
            self.exit_on_failure(f"Invalid SCORE source - '{src}'", force_exit=True)
            return

        _content = gen_deploy_data_content(src)
        _request_payload = self._convert_valid_payload_format(
            method="icx_sendTransaction",
            params={
                "to": self.governance_address,
                "dataType": "deploy",
                "data": {
                    "contentType": f"application/{_file_type}",
                    "content": f'0x{_content.hex()}',
                    "params": params,
                },
            },
        )
        self.global_reqeust_payload = _request_payload
        return _request_payload

    def deploy_score(self, src="", params={}, step_limit=None, governance_address=None, is_wait=True, is_confirm_send=False):
        payload = self.create_deploy_payload(src, params, governance_address)
        self.sign_tx(payload=payload, step_limit=step_limit)

        if is_confirm_send:
            print_json(self.signed_tx)
            is_send = Confirm.ask("Do you want to send the [bold red]deploy[/bold red] transaction?", default=False)
        else:
            is_send = True

        if is_send:
            response = self.sign_send(is_wait=is_wait)
            return response

    def get_step_price(self, ):
        response = self.governance_call(method="getStepPrice", governance_address=const.CHAIN_SCORE_ADDRESS, sign=False, store_request_payload=False)
        return response

    def _get_step_costs(self, url=None):
        response = self.governance_call(method="getStepCosts", governance_address=const.CHAIN_SCORE_ADDRESS, sign=False, store_request_payload=False)
        return response

    def get_step_cost(self, step_kind="apiCall"):
        return self._get_step_costs().get(step_kind)

    def get_estimate_step(self, url=None, tx=None):
        if url:
            _url = url
        else:
            _url = self.debug_url

        _tx = copy.deepcopy(tx)
        _tx = self.auto_fill_parameter(_tx)

        if isinstance(_tx, dict):
            _tx['method'] = "debug_estimateStep"
            res = self.rpc_call(
                url=_url,
                payload=_tx,
                store_request_payload=False,
                print_error=True
            )
            res_json = res
            if res_json.get('error'):
                # pawn.console.debug(f"[red] An error occurred while running debug_estimateStep, {res_json['error'].get('message')}")
                # sys.exit(-1)
                self.exit_on_failure(f"An error occurred while running debug_estimateStep, {res_json['error'].get('message')}")

            return res.get('result')

        else:
            self.exit_on_failure(f"TX is not dict. tx => {_tx}")

    def get_step_limit(self, url=None, tx=None, step_kind="apiCall"):
        if tx:
            _tx = tx
        else:
            _tx = copy.deepcopy(self.request_payload)
        unnecessary_keys = ["stepLimit", "signature"]
        for key in unnecessary_keys:
            if _tx['params'].get(key, '__NOT_DEFINED__') != "__NOT_DEFINED__":
                pawn.console.debug(f"Remove the unnecessary '{key}' in payload")
                del _tx['params'][key]

        estimate_step = self.get_estimate_step(tx=_tx)
        _guess_step_kind = self._guess_step_kind(tx=_tx)
        pawn.console.debug(f"guess step_kind = {_guess_step_kind}")
        step_cost = self.get_step_cost(_guess_step_kind)
        step_price = self.get_step_price()

        if estimate_step and step_cost:
            step_limit = hex(hex_to_number(estimate_step) + hex_to_number(step_cost))
            icx_fee = hex_to_number(estimate_step) * hex_to_number(step_price) / const.TINT
            pawn.console.debug(f"fee = {icx_fee} => estimate[i]({hex_to_number(estimate_step, debug=True)})[/i] * "
                               f"step_price[i]({hex_to_number(step_price, debug=True)})[/i]")

            pawn.console.debug(f"step_limit => {hex_to_number(step_limit, debug=True)}")
            return step_limit
        else:
            _default_step_limit = self.default.get("stepLimit")
            pawn.console.log(f"[red][FAIL][/red] An error occurred while running get_step_limit(). set default : {_default_step_limit}")
            return _default_step_limit

    @staticmethod
    def _guess_step_kind(tx=None):
        step_kind_dict = {
            "dataType": {
                "deploy": "contractCreate",
                "call": "apiCall"
            }
        }
        _default_step_kind = "get"
        if not isinstance(tx, dict):
            return _default_step_kind

        for payload_key, value in step_kind_dict.items():
            if keys_exists(tx, "params", payload_key):
                _data_type = tx['params'][payload_key]
                return step_kind_dict[payload_key].get(_data_type, _default_step_kind)
        return _default_step_kind

    def get_fee(self, tx=None, symbol=False):
        _tx = self.parse_tx_var(tx)
        if keys_exists(_tx, "params", "stepLimit"):
            del _tx['params']['stepLimit']
        estimate_step = self.get_estimate_step(tx=_tx)
        step_kind = self._guess_step_kind(tx)
        pawn.console.debug(f"_guess_step_kind = {step_kind}")
        step_cost = self.get_step_cost(step_kind)
        step_price = self.get_step_price()
        step_limit = hex(hex_to_number(estimate_step) + hex_to_number(step_cost))

        fee = hex_to_number(estimate_step) * hex_to_number(step_price) / const.TINT
        pawn.console.debug(f"[red] fee = (estimate_step + step_cost) * step_price = {fee}")
        pawn.console.debug(f"[red] fee = ({estimate_step} + {step_cost}) * {step_price} = {fee}")

        if symbol:
            fee = f"{fee} {self.network_info.symbol}"

        return fee

    def get_score_api(self, address="", url=None):
        if not is_valid_token_address(address, prefix="cx"):
            return self.exit_on_failure(f"Invalid token address - {address}")

        response = self.rpc_call(
            url=url,
            method="icx_getScoreApi",
            params={"address": address},
            store_request_payload=False,
        )

        error = self.response.get('error')
        if error:
            self.exit_on_failure(error)
            return 0

        return response.get('result')

    @staticmethod
    def name_to_params(list_data):
        return {data.get('name'): "" for data in list_data}

    def _convert_score_api_to_params(self, data=[], address=""):
        result = {}
        for _input in data:
            if _input.get('type') != "function":
                continue

            method = "icx_call" if _input.get('readonly') == "0x1" else "icx_sendTransaction"
            score_method = _input.get('name')

            result[score_method] = dict(
                method=method,
                params=dict(
                    dataType="call",
                    to=address,
                    data=dict(
                        method=score_method,
                        params=self.name_to_params(_input.get('inputs'))
                    )
                )
            )

            if method == "icx_sendTransaction":
                result[score_method]['params']['value'] = "0x0"
            # pawn.console.log(f"{_input.get('type')} {score_method} , {_input.get('inputs')}, {_input.get('readonly')}")
        return result

    def get_governance_api(self, url=None):
        self.score_api = {}
        for address in [const.GOVERNANCE_ADDRESS, const.CHAIN_SCORE_ADDRESS]:
            _api_result = self.get_score_api(address=address, url=url)
            if _api_result:
                self.score_api[address] = self._convert_score_api_to_params(_api_result, address=address)

        return self.score_api

    def get_balance(self, url=None, address=None, is_comma=False):
        if not address and self.wallet:
            address = self.wallet.get('address')

        if is_valid_token_address(address):
            response = self.rpc_call(
                url=url,
                method="icx_getBalance",
                params={"address": address},
                store_request_payload=False,
            )
            _balance = response.get('result')
            if self.response.get('error'):
                self.exit_on_failure(self.response.get('error'))
                return 0
            return hex_to_number(_balance, is_comma=is_comma, is_tint=True)
        else:
            return self.exit_on_failure(f"Invalid token address - {address}")

    def get_tx(self,  tx_hash=None, url=None,  return_key=None):
        if not tx_hash:
            tx_hash = self._get_tx_hash
        if not is_valid_tx_hash(tx_hash):
            self.exit_on_failure(f"Invalid tx_hash - {tx_hash}")

        response = self.rpc_call(
            url=url,
            method="icx_getTransactionResult",
            params={"txHash": tx_hash}
        )
        if isinstance(response, dict):
            if return_key:
                return response.get(return_key)
            return response
        return response.get('text')

    def get_tx_wait(self,  tx_hash=None, url=None,  is_compact=True):
        tx_hash = self._get_tx_hash(tx_hash)
        if not tx_hash:
            return
        self.on_error = False
        count = 0
        with pawn.console.status("[magenta] Wait for transaction to be generated.") as status:
            while True:
                resp = self._check_transaction(url, tx_hash)
                if resp.get('error'):
                    text, exit_loop = self._handle_error_response(resp, count, tx_hash)
                elif resp.get('result'):
                    text, exit_loop = self._handle_success_response(resp, count, tx_hash)
                else:
                    text = resp
                status.update(
                    status=text,
                    spinner_style="yellow",
                )
                if exit_loop:
                    self._print_final_result(text, is_compact, resp)
                    break
                count += 1
                time.sleep(1)
            if self.on_error and tx_hash:
                self.get_debug_trace(tx_hash, reset_error=False)
        return resp

    def _get_tx_hash(self, tx_hash):
        if not tx_hash and isinstance(self.response, dict):
            if keys_exists(self.response, 'json', 'result', 'txHash'):
                tx_hash = self.response['json']['result']['txHash']
            elif keys_exists(self.response, 'json', 'result'):
                tx_hash = self.response['json']['result']
        if not tx_hash:
            pawn.console.log(f"[red] Not found tx_hash='{tx_hash}'")
        return tx_hash

    def _check_transaction(self, url, tx_hash):
        return self.get_tx(url=url, tx_hash=tx_hash)

    def _handle_error_response(self, resp, count, tx_hash):
        exit_loop = False
        exit_msg = ""
        prefix_text = f"[cyan][Wait TX][{count}][/cyan] Check a transaction by [i cyan]{tx_hash}[/i cyan] "
        _error_message = resp['error'].get('message')
        exit_error_messages = ["InvalidParams", "NotFound: E1005:not found tx id"]

        if any(error_message in _error_message for error_message in exit_error_messages):
            exit_loop = True
            exit_msg = "[red][FAIL][/red]"
            self.on_error = True
        text = f"{prefix_text}{exit_msg}[white] '{_error_message}'"
        return text, exit_loop

    def _handle_success_response(self, resp, count, tx_hash):
        if resp['result'].get('logsBloom'):
            resp['result']['logsBloom'] = int(resp['result']['logsBloom'], 16)
        if resp['result'].get('failure'):
            _resp_status = "[red][FAIL][/red]"
            self.on_error = True
        else:
            _resp_status = "[green][OK][/green]"
            self.on_error = False
        exit_loop = True
        prefix_text = f"[cyan][Wait TX][{count}][/cyan] Check a transaction by [i cyan]{tx_hash}[/i cyan] "
        text = align_text(prefix_text, _resp_status, ".")
        return text, exit_loop

    def _print_final_result(self, text, is_compact, resp):
        final_result = f"[bold green][white]{text}"
        if is_compact:
            final_result = shorten_text(final_result, width=205)
        pawn.console.tprint(final_result)
        if self.debug or pawn.get('PAWN_DEBUG'):
            print_json(resp['result'])

    def get_debug_trace(self, tx_hash, print_table=True, reset_error=True):
        response = self.rpc_call(
            url=self.debug_url,
            method="debug_getTrace",
            params={"txHash": tx_hash},
            store_request_payload=False,
            reset_error=reset_error,
        )
        if isinstance(response, dict) and keys_exists(response, "result", "logs") and print_table:
            PrintRichTable(f"debug_getTrace={tx_hash}", data=response['result']['logs'])
        else:
            pawn.console.log(response)
        return response

    def parse_tx_var(self, tx=None):
        if tx:
            _tx = copy.deepcopy(tx)
            self._use_global_reqeust_payload = False
        else:
            _tx = self.request_payload
            self._use_global_reqeust_payload = True
            # pawn.console.debug(f"[red]_use_global_reqeust_payload={self._use_global_reqeust_payload}")
        return _tx

    def _force_fill_from_address(self, tx=None):
        _tx = self.parse_tx_var(tx)
        if not isinstance(_tx, dict):
            pawn.console.debug(f"tx is not dict, tx={tx}, _tx={_tx}, request_payload = {self.request_payload}")
            exit()

        if self.wallet and self.wallet.get('address'):
            _tx['params']['from'] = self.wallet.get('address')
        else:
            self.exit_on_failure(exception="Not found 'from address'. Not defined wallet")
        return _tx

    def auto_fill_parameter(self, tx=None, is_force_from_addr=False):
        _tx = self.parse_tx_var(tx)
        if isinstance(_tx, dict) and _tx.get('params'):
            if not _tx['params'].get('from') or is_force_from_addr:
                self._force_fill_from_address()
            if not _tx['params'].get('nonce'):
                _tx['params']['nonce'] = "0x1"

            if not _tx['params'].get('version'):
                _tx['params']['version'] = "0x3"

            if not _tx['params'].get('timestamp'):
                _tx['params']['timestamp'] = hex(icx_signer.get_timestamp_us())

            if not _tx['params'].get('nid') and self.network_info and self.network_info.nid:
                _tx['params']['nid'] = self.network_info.nid
        else:
            self.exit_on_failure(f"Invalid payload => {_tx}")

        if self._use_global_reqeust_payload:
            self.request_payload = _tx
        return _tx

    def auto_fill_step_limit(self, tx=None, step_limit=None):
        _tx = self.parse_tx_var(tx)
        if isinstance(_tx, dict) and _tx.get('params'):
            if step_limit:
                if not is_hex(step_limit):
                    pawn.console.debug(f"Converting step limit to hex {step_limit}=>{hex(int(step_limit))}")
                    step_limit = hex(int(step_limit))

                _tx['params']['stepLimit'] = step_limit
            elif not _tx['params'].get('stepLimit'):
                _tx['params']['stepLimit'] = self.get_step_limit()
        else:
            self.exit_on_failure(f"Invalid payload => {_tx}")

    def sign_tx(self, wallet=None, payload=None, check_balance=True, step_limit=None):
        self.request_payload = {}
        self.signed_tx = {}
        if wallet:
            self.wallet = wallet

        if not self.wallet or not isinstance(self.wallet, dict):
            self.exit_on_failure(f"[red] Not defined wallet => {self.wallet}")

        if not payload and self.global_reqeust_payload:
            pawn.console.debug("Use global_reqeust_payload")
            payload = self.global_reqeust_payload

        self.request_payload = self._convert_valid_payload_format(payload=payload)
        private_key = self.wallet.get('private_key')
        address = self.wallet.get('address')

        if check_balance:
            _balance = self.get_balance()
            if self.network_info:
                symbol = self.network_info.symbol
            else:
                symbol = ""
            pawn.console.log(f"<{address}>'s Balance = {_balance} {symbol}")

            if not _balance or _balance == 0:
                return self.exit_on_failure(f"<{address}> Out of Balance = {_balance} {symbol}")
        self.auto_fill_parameter(is_force_from_addr=True)
        self.auto_fill_step_limit(step_limit=step_limit)

        singer = icx_signer.IcxSigner(data=private_key)
        self.signed_tx = singer.sign_tx(self.request_payload)

        if address != singer.get_hx_address():
            self.exit_on_failure(f'Invalid address {address} != {singer.get_hx_address()}')

        self.global_reqeust_payload = {}
        return self.signed_tx

    def exit_on_failure(self, exception, force_exit=False):
        self.on_error = True
        if self.raise_on_failure:
            raise NoTraceBackException(exception)
        else:
            pawn.console.log(f"[red][FAIL][/red] Stopped {get_debug_here_info().get('function_name')}(), {exception}")
            if force_exit:
                sys_exit()
        return exception

    def sign_send(self, is_wait=True, is_compact=False):
        if self.signed_tx:
            response = self.rpc_call(payload=self.signed_tx, print_error=True)
            if not is_wait:
                return response

            if isinstance(response, dict) and response.get('result'):
                resp = self.get_tx_wait(tx_hash=response['result'], is_compact=is_compact)
                return resp
        else:
            self.exit_on_failure(f"Required signed transaction")

    @staticmethod
    def _check_tx_result(result):
        if result:
            # print(result)
            response_json = result
            if response_json.get("error"):
                return False
            else:
                if response_json.get('result'):
                    return True
                return response_json


class JsonRequest:
    def __init__(self):
        """
        TODO: It will be generated the JSON or JSON-RPC request
        """
        pass


def disable_ssl_warnings():
    from urllib3.exceptions import InsecureRequestWarning
    requests.packages.urllib3.disable_warnings(InsecureRequestWarning)


def append_http(url):
    """

    Add http:// if it doesn't exist in the URL

    :param url:
    :return:
    """

    if "http://" not in url and "https://" not in url:
        url = f"http://{url}"
    return url


def append_ws(url):
    """

    Add ws:// if it doesn't exist in the URL

    :param url:
    :return:
    """
    if "https://" in url:
        url = url.replace("https://", "wss://")
    elif "http://" in url:
        url = url.replace("http://", "ws://")
    elif "ws://" not in url and "wss://" not in url:
        url = f"ws://{url}"
    return url


def append_api_v3(url):
    if "/api/v3" not in url:
        return f"{url}/api/v3"
    return append_http(url)


def remove_http(url):
    """
    Remove the r'https?://' string

    :param url:
    :return:

    """
    return re.sub(r"https?://", '', url)


def get_operator_truth(inp, relate, cut):
    ops = {
        '>': _operator.gt,
        '<': _operator.lt,
        '>=': _operator.ge,
        '<=': _operator.le,
        '==': _operator.eq,
        '!=': _operator.ne,
        'include': lambda y, x: x in y,
        'exclude': lambda y, x: x not in y
    }
    return ops[relate](inp, cut)


def guess_key(find_key, data):
    guessed_result = []
    if isinstance(data, dict):
        for k, v in data.keys():
            if find_key in k:
                guessed_result.append(k)
    elif isinstance(data, list):
        for k in data:
            if find_key in k:
                guessed_result.append(k)
    return guessed_result


class SuccessCriteria:
    def __init__(
            self,
            target: str = "",
            operator: Literal["!=", "==", ">=", "<=", ">", "<", "include", "exclude"] = "",
            expected: Union[str, int, float] = "",
            # operator: Literal[tuple(ALLOW_OPERATOR)] = ""
    ):
        if target:
            self.target = target
        self.operator = operator
        self.expected = expected
        self.result = False

        if (self.expected or self.expected == 0) and self.target and self.operator:
            try:
                self._set_string_valid_type()
                self.result = get_operator_truth(self.target, self.operator, self.expected)
                pawn.console.debug(f"get_operator_truth(target={self.target}, {self.operator}, expected={self.expected}) = {self.result}")
            except:
                self.result = False

    def _set_string_valid_type(self, ):
        _debug_message = ""
        for var_name in ["target", "expected"]:
            _attr_value_in_class = getattr(self, var_name, "__NOT_NONE__")
            if _attr_value_in_class != "__NOT_NONE__":
                if isinstance(_attr_value_in_class, str):
                    if is_int(_attr_value_in_class):
                        setattr(self, var_name, int(_attr_value_in_class))
                    elif is_float(_attr_value_in_class):
                        setattr(self, var_name, float(_attr_value_in_class))

                _modified_value = getattr(self, var_name)
                _debug_message += f"{var_name} = {_modified_value} ({type(_modified_value)}) , "

        # pawn.console.debug(_debug_message)

    def __str__(self):
        return "<SuccessCriteria %s>" % self.__dict__

    def __repr__(self):
        return "<SuccessCriteria %s>" % self.__dict__

    def to_dict(self):
        return self.__dict__


class SuccessResponse(SuccessCriteria):
    def __init__(
            self,
            target_key: str = "",
            operator: Literal["!=", "==", ">=", "<=", ">", "<", "include", "exclude"] = "",
            expected: Union[str, int, float] = "",
            target: dict = {},
    ):
        # if not target or not operator or not expected or target_key:
        #     raise ValueError(f"target: {target}, operator: {operator}, expected: {expected}, target_key: {target_key} ")

        if not isinstance(target, dict):
            pawn.console.log(f"[red]<Error>[/red] '{target}' is not dict")
            self.result = False
            raise ValueError(f"target is not dict - '{target}'")

        self.target_key = target_key
        self.target = FlatDict(target)

        _selected_flatten_target = self.target.get(self.target_key)
        super().__init__(target=_selected_flatten_target, operator=operator, expected=expected)

        if not _selected_flatten_target:
            pawn.console.debug(f"[red]<Error>[/red] '{self.target_key}' is not attribute in {list(self.target.keys())}")
            pawn.console.debug(
                f"[red]<Error>[/red] '{self.target_key}' not found. \n Did you mean {guess_key(self.target_key, self.target.keys())} ?")
            self.result = False


class CallHttp:
    SHORTEN_MESSAGE_DICT = {
        requests.exceptions.Timeout: {
            "message": "Timeout Error",
            "params_message": "timeout={}"
        },
        requests.exceptions.HTTPError: "HTTP Error",
        requests.exceptions.ConnectionError: "DNS lookup Error",
        requests.exceptions.RequestException: "OOps: Something Else",
    }

    def __init__(self,
                 url=None,
                 method: Literal["get", "post", "patch", "delete"] = "get",
                 # method: Literal[AllowsHttpMethod.get] = "get",
                 # method: AllowsHttpMethod = AllowsHttpMethod.get,
                 # method: Literal[tuple(method for method in AllowsHttpMethod)],
                 payload={},
                 timeout=3000,
                 ignore_ssl: bool = False,
                 verbose: int = 0,
                 success_criteria: Union[dict, list, str, None] = "__DEFAULT__",
                 success_operator: Literal["and", "or"] = "and",
                 success_syntax: Literal["operator", "string", "auto"] = "auto",
                 raise_on_failure: bool = False,
                 auto_run: bool = True,
                 **kwargs
                 ):

        self.url = url
        self.method = method.lower()
        self.payload = payload
        self.timeout = timeout / 1000
        self.ignore_ssl = ignore_ssl
        self.verbose = verbose

        _default_criteria = [AllowsKey.status_code, "<=", 399]
        if success_criteria == '__DEFAULT__':
            self.success_criteria = _default_criteria
        elif success_criteria:
            self.success_criteria = success_criteria
        else:
            self.success_criteria = None

        self.success_operator = success_operator
        self.success_syntax = success_syntax
        self.kwargs = kwargs
        self.raise_on_failure = raise_on_failure
        self._DEFAULT_UA = f"CallHttp Agent/{pawn.get('PAWN_VERSION')}"
        self.on_error = False
        # self.response = requests.models.Response()
        self.response = HttpResponse()
        self.flat_response = None

        self.success = None
        self._success_results = []
        self._success_criteria = None
        self.timing = 0
        # self.run()

    def _shorten_exception_message_handler(self, exception):
        default_msg = f"(url={self.url} method={self.method.upper()}"
        for req_exception, values in self.SHORTEN_MESSAGE_DICT.items():
            if isinstance(exception, req_exception):
                if isinstance(values, dict) and values.get('message'):
                    _message = f"<{values.get('message')}>"
                    _params_message = f" {values.get('params_message', '')}"
                else:
                    _message = f"<{values}>"
                    _params_message = ''
                return f"{_message}{default_msg}{_params_message})"

        return f"<Unknown Error>{default_msg})"

    def exit_on_failure(self, exception):
        self.on_error = True
        self.response = HttpResponse(status_code=999, error=self._shorten_exception_message_handler(exception), elapsed=self.timing)
        if self.raise_on_failure:
            raise NoTraceBackException(exception)
        else:
            pawn.console.debug(f"[red][FAIL][/red] {exception}")
            # self.response.status_code = 999
            # self.response.success = False
            # self.response.error = self._shorten_exception_message_handler(exception)
            # print(self.timing)
            return self.response

    def run(self):
        self._prepare()
        start = time.perf_counter()
        self.fetch_response()
        end = time.perf_counter()
        self.timing = int((end - start) * 1000)
        self._parse_response()
        self.fetch_criteria()
        self.response.success = self.is_success()

        if not self.is_success() and getattr(self.response, 'reason', ''):
            self.response.error = f"{self.response.status_code} {self.response.reason}"

        return self

    def _prepare(self):
        self.url = append_http(self.url)

    def get_response(self) -> HttpResponse:
        return self.response

    def fetch_response(self):
        (json_response, data, http_version, r_headers, error) = ({}, {}, None, None, None)
        if self.method not in ("get", "post", "patch", "delete"):
            pawn.error_logger.error(f"unsupported method='{self.method}', url='{self.url}' ") if pawn.error_logger else False
            return self.exit_on_failure(f"Unsupported method={self.method}, url={self.url}")
        try:
            try:
                _payload_string = json.dumps(self.payload)
            except Exception as e:
                _payload_string = self.payload
            pawn.console.debug(f"[TRY] url={self.url}, method={self.method.upper()}, kwargs={self.kwargs}")
            if pawn.get("PAWN_DEBUG"):
                print_json(_payload_string)

            func = getattr(requests, self.method)
            if self.method == "get":
                self.response = func(self.url, verify=False, timeout=self.timeout, **self.kwargs)
            else:
                self.response = func(self.url, json=self.payload, verify=False, timeout=self.timeout, **self.kwargs)
        except Exception as e:
            return self.exit_on_failure(e)

    def _parse_response(self):
        self.response.timing = self.timing
        try:
            _elapsed = int(self.response.elapsed.total_seconds() * 1000)
        except AttributeError:
            if self.timing:
                _elapsed = self.timing
            else:
                _elapsed = 0
        self.response.elapsed = _elapsed

        if getattr(self.response, 'raw', None):
            self.response.http_version = self.response.raw.version
        else:
            self.response.http_version = ""
        if self.response and not self.on_error:
            try:
                self.response.result = self.response.json()
            except:
                self.response.result = self.response.text

    def fetch_criteria(self,
                       success_criteria: Union[dict, list] = None,
                       success_operator: Literal["and", "or"] = "and",
                       ):
        _response_dict = self.response.as_dict()

        if success_criteria:
            self.success_criteria = success_criteria
        if success_operator:
            self.success_operator = success_operator

        if not self.success_criteria:
            pawn.console.debug("passing success_criteria")
        else:
            if self.success_syntax == "string" or self.success_syntax == "auto":
                _check_syntax = self._check_criteria_syntax()
                if _check_syntax:
                    pawn.console.debug(f"[blue]Try to convert[/blue] {type(self.success_criteria)}, _check_criteria_syntax={_check_syntax}")
                    self._recursive_convert_criteria()

            depth = list_depth(self.success_criteria)
            if depth == 1:
                self.success_criteria = [self.success_criteria]

            for criteria in self.success_criteria:
                # pawn.console.debug(f"Compare Criteria => {type(criteria)} {criteria}")
                if isinstance(criteria, list):
                    _criteria = copy.deepcopy(criteria)
                    _criteria.append(_response_dict)
                    self._success_results.append(SuccessResponse(*_criteria))
                elif isinstance(criteria, dict):
                    criteria['target'] = _response_dict
                    self._success_results.append(SuccessResponse(**criteria))
            # pawn.console.debug(self._success_results)

    @staticmethod
    def _find_operator(string):
        for operator in ALLOW_OPERATOR:
            if operator in string:
                return True
        return False

    def _check_criteria_syntax(self, criteria=None):
        if not criteria:
            criteria = self.success_criteria

        if isinstance(criteria, str):
            if self._find_operator(criteria):
                return True
            return False

        elif isinstance(criteria, list):
            for ct in criteria:
                return self._check_criteria_syntax(ct)

    def _recursive_convert_criteria(self, string_criteria=None, depth=None):

        if not string_criteria:
            string_criteria = self.success_criteria

        if string_criteria and isinstance(string_criteria, str):
            self.success_criteria = self._convert_criteria(string_criteria)
            return self.success_criteria

        elif isinstance(self.success_criteria, list):
            string_criteria_in_list = []
            for string_criteria in string_criteria:
                converted_criteria = self._recursive_convert_criteria(string_criteria)
                if converted_criteria:
                    string_criteria_in_list.append(converted_criteria)

            if len(string_criteria_in_list) > 0:
                self.success_criteria = string_criteria_in_list

    def _convert_criteria(self, argument):
        for operator in ALLOW_OPERATOR:
            if operator in argument:
                result = argument.split(operator)
                if any(word in result[0] for word in ALLOW_OPERATOR + ['=']):
                    pawn.console.log(f"[red]Invalid operator - '{argument}', {result}")
                    raise ValueError(f"Invalid operator - '{argument}', {result}")
                result.insert(1, operator)
                return result
        return False

    def _convert_list_criteria(self, arguments):
        result = []
        for argument in arguments:
            criteria = self._convert_criteria(argument)
            if criteria:
                result.append(criteria)
        return result

    def is_success(self):
        success_count = 0
        expected_count = len(self._success_results)

        if isinstance(self._success_results, list):
            for _result in self._success_results:
                if _result.result:
                    success_count += 1

        if self.success_operator == "and" and success_count == expected_count:
            return True
        elif self.success_operator == "or" and success_count > 0:
            return True
        return False


class CheckSSL:

    def __init__(self, host=None, timeout: float = 5.0, port: int = 443, raw_data=False):
        self._host = host
        self._timeout = timeout
        self._port = port
        self.ssl_info = None
        self._now = datetime.now()
        self._raw_data = raw_data
        self.ssl_dateformat = r'%b %d %H:%M:%S %Y %Z'
        self.get_ssl()
        self.ssl_result = {}

    def get_ssl(self):
        context = ssl.create_default_context()
        context.check_hostname = False

        conn = context.wrap_socket(
            socket.socket(socket.AF_INET),
            server_hostname=self._host,
        )
        conn.settimeout(self._timeout)
        conn.connect((self._host, self._port))
        conn.do_handshake()
        self.ssl_info = conn.getpeercert()
        self._parse_ssl_info()
        conn.close()
        return self.ssl_info

    def _parse_ssl_info(self):
        expire_date = self.ssl_expiry_datetime()
        diff = expire_date - datetime.now()
        self._calculated_date = {
            "expire_date": expire_date.strftime("%Y-%m-%d"),
            "left_days": diff.days
        }
        self.ssl_info.update(self._calculated_date)
        if not self._raw_data:
            _tmp_ssl_info = {}
            _to_dict_keys = ["subject", "issuer"]
            _to_list_keys = ["subjectAltName"]

            for key, value in self.ssl_info.items():
                if key in _to_dict_keys:
                    self.ssl_info[key] = self._tuple_to_dict(value)
                elif key in _to_list_keys:
                    print(f"key={key}, value={value}")
                    self.ssl_info[key] = self._tuple_to_list(value)

    @staticmethod
    def _tuple_to_dict(_tuple=None):
        _tmp_dict = {}
        if isinstance(_tuple, tuple):
            for _data in _tuple:
                _tmp_dict.update(_data)
        return _tmp_dict

    @staticmethod
    def _tuple_to_list(_tuple=None):
        _tmp_dict = {}
        if isinstance(_tuple, tuple):
            for _data in _tuple:
                _key = _data[0]
                _domain = _data[1]
                pawn.console.log(f"[red] _tuple_to_list ={_key}, {_domain}")
                if not _tmp_dict.get(_key):
                    _tmp_dict[_key] = []
                _tmp_dict[_key].append(_domain)
        return _tmp_dict

    def ssl_expiry_datetime(self) -> datetime:
        return datetime.strptime(self.ssl_info['notAfter'], self.ssl_dateformat)

    def analyze_ssl(self):
        api_url = 'https://api.ssllabs.com/api/v3/'
        _do_function = partial(jequest, f"{api_url}/analyze?host={self._host}")
        WaitStateLoop(
            loop_function=_do_function,
            exit_function=self._check_ssl_result,
        ).run()

    def _check_ssl_result(self, result):
        if result.get('json') and result['json'].get('status') == "READY":
            self.ssl_result = result['json']
            pawn.console.log(self.ssl_result)
            return True


class CallWebsocket:

    def __init__(
            self,
            connect_url,
            verbose=0,
            timeout=10,
            send_callback: Callable[..., Any] = None,
            recv_callback: Callable[..., Any] = None,
            use_status_console: bool = False,
    ):
        self.connect_url = connect_url
        self.verbose = verbose
        self.timeout = timeout
        self.send_callback = send_callback
        self.recv_callback = recv_callback

        self.ws_url = append_ws(connect_url)
        self.http_url = append_http(connect_url)
        self.status_console = Null()
        self._use_status_console = use_status_console

        self._ws = None

        if self.verbose > 3:
            enableTrace(True)

    def connect_websocket(self, api_url=""):
        self._ws = create_connection(f"{self.ws_url}/{api_url}", timeout=self.timeout, sslopt={"cert_reqs": ssl.CERT_NONE})
        self._ws.settimeout(self.timeout)

    def run(self, api_url="api/v3/icon_dex/block", status_console=False):
        self.connect_websocket(api_url)
        if self._use_status_console or status_console:
            with pawn.console.status("Call WebSocket") as self.status_console:
                self.send_recv_callback()
        else:
            self.send_recv_callback()

    def send_recv_callback(self):
        if callable(self.send_callback):
            self._ws.send(self.send_callback())

        while True:
            response = self._ws.recv()
            if callable(self.recv_callback):
                self.recv_callback(response)


class GoloopWebsocket(CallWebsocket):

    def __init__(self,
                 connect_url,
                 verbose=0,
                 timeout=10,
                 blockheight=0,
                 sec_thresholds=4,
                 monitoring_target=None,
                 ignore_ssl=True,
                 ):

        self.connect_url = connect_url
        self.verbose = verbose
        self.timeout = timeout
        self.blockheight = blockheight
        self.sec_thresholds = sec_thresholds
        self.monitoring_target = monitoring_target
        if self.monitoring_target is None:
            self.monitoring_target = ["block"]

        self.compare_diff_time = {}
        self.delay_cnt = {}

        self.block_timestamp_prev = 0
        self.block_timestamp = None
        self.tx_count = 0
        self.tx_timestamp = 0
        self.tx_timestamp_dt = None

        self.blockheight_now = 0

        if ignore_ssl:
            disable_ssl_warnings()

        if self.verbose > 0:
            use_status_console = True
        else:
            use_status_console = False

        super().__init__(
            connect_url=self.connect_url,
            verbose=self.verbose,
            timeout=self.timeout,
            send_callback=self.request_blockheight_callback,
            recv_callback=self.parse_blockheight,
            use_status_console=use_status_console
        )

    def request_blockheight_callback(self):
        if self.blockheight == 0:
            self.blockheight = self.get_last_blockheight()
        pawn.console.log(f"Call request_blockheight_callback - blockheight: {self.blockheight:,}")
        send_data = {
            "height": hex(self.blockheight)
        }
        return json.dumps(send_data)

    def parse_blockheight(self, response=None):
        response_json = json.loads(response)
        self.compare_diff_time = {}

        if response_json and response_json.get('hash'):
            hash_result = self.get_block_hash(response_json.get('hash'))
            self.blockheight_now = hash_result.get("height")
            pawn.set(LAST_EXECUTE_POINT=self.blockheight_now)
            self.block_timestamp = hash_result.get("time_stamp")

            _message = f"[bold][📦 {self.blockheight_now:,}][/bold] 📅 {date_utils.timestamp_to_string(self.block_timestamp)}, tx_hash: {hash_result.get('block_hash')}"
            pawn.console.debug(_message)
            self.status_console.update(_message)

            if self.block_timestamp_prev != 0:
                self.compare_diff_time['block'] = abs(self.block_timestamp_prev - self.block_timestamp)

            tx_list = hash_result.get('confirmed_transaction_list')
            self.tx_count = len(tx_list)

            for tx in tx_list:
                self.tx_timestamp = int(tx.get("timestamp", "0x0"), 0)  # 16진수, timestamp가 string이여야한다.
                if self.tx_timestamp:
                    self.compare_diff_time['tx'] = abs(self.block_timestamp - self.tx_timestamp)

                for target in self.monitoring_target:
                    diff_time = self.compare_diff_time.get(target, 0) / 1_000_000
                    if diff_time != 0 and abs(diff_time) > self.sec_thresholds:
                        pawn.console.log(f"[{target.upper():^5}] {self.output_format(key_string=target, diff_time=diff_time, is_string=True)}")
                        if self.verbose > 2:
                            dump(hash_result['confirmed_transaction_list'])
            self.block_timestamp_prev = self.block_timestamp

        else:
            pawn.console.log(response_json)

    def output_format(self, key_string="", diff_time=0, is_string=False):
        self.delay_cnt[key_string] = self.delay_cnt.get(key_string, 0) + 1

        blockheight_date = date_utils.timestamp_to_string(self.block_timestamp)
        if key_string == "block":
            diff_message = f"BH_time(now): {blockheight_date}, " \
                           f"BH_time(prev): {date_utils.timestamp_to_string(self.block_timestamp_prev)}"
        else:
            diff_message = f"BH_time: {blockheight_date}, TX_time: {date_utils.timestamp_to_string(self.tx_timestamp)}"

        result = (f"<{self.delay_cnt[key_string]}> [{key_string} Delay][{date_utils.second_to_dayhhmm(diff_time)}] ",
                  f"BH: {self.blockheight_now}, "
                  # f"diff: {date_utils.second_to_dayhhmm(diff_time)}, "
                  f"TX_CNT:{self.tx_count}, {diff_message}")
        if is_string:
            return "".join(result)
        return result

    def get_last_blockheight(self):
        res = jequest(method="post", url=f"{self.http_url}/api/v3", data=generate_json_rpc(method="icx_getLastBlock"))
        pawn.console.log(res['json'].get('result'))
        if res['json'].get('result'):
            return res['json']['result'].get('height')

    def get_block_hash(self, hash):
        res = jequest(
            method="post",
            url=f"{self.http_url}/api/v3",
            data=generate_json_rpc(method="icx_getBlockByHash", params={"hash": hash})
        )
        if res.get('json'):
            return res['json'].get('result')
        else:
            pawn.console.log(f"[red] {res}")


def gen_rpc_params(method=None, params=None):
    default_rpc = {
        "jsonrpc": "2.0",
        "id": 1234,
        "method": method,
    }

    if params:
        default_rpc['params'] = params
    return default_rpc


def getBlockByHash(nodeHost, hash):
    url = append_http(f"{nodeHost}/api/v3")
    data = gen_rpc_params(method="icx_getBlockByHash", params={"hash": hash})
    response = requests.post(url=url, data=json.dumps(data), timeout=10, verify=False)
    return response.json()


def getTransactionByHash(nodeHost, hash):
    url = append_http(f"{nodeHost}/api/v3")
    data = gen_rpc_params(method="icx_getTransactionByHash", params={"txHash": hash})
    response = requests.post(url=url, data=json.dumps(data), timeout=10, verify=False)
    return response.json()


def getLastBlock(nodeHost):
    url = append_http(f"{nodeHost}/api/v3")
    data = gen_rpc_params(method="icx_getLastBlock")
    response = requests.post(url=url, data=json.dumps(data), timeout=10, verify=False)

    if response.status_code == 200:
        json_res = response.json()
        if json_res.get("result"):
            return json_res["result"]["height"]
    else:
        print(response)
        sys.exit()
    return 0


def jequest(url, method="get", payload={}, elapsed=False, print_error=False, timeout=None, ipaddr=None, **kwargs) -> dict:
    """
    This functions will be called the http requests.

    :param url:
    :param method:
    :param payload:
    :param elapsed:
    :param print_error:
    :param timeout: Timeout seconds
    :param ipaddr: Change the request IP address in http request
    :param \*\*kwargs: Optional arguments that ``request`` takes.

    :return:
    """
    if ipaddr:
        if url.startswith('http') or url.startswith('http://'):
            domain = re.sub(r'https?://', '', url)
            # net.override_dns(domain, ipaddr)
            # net.override_dns(domain=domain, ipaddr=ipaddr)
            net.OverrideDNS(domain=domain, ipaddr=ipaddr).set()
    else:
        net.OverrideDNS().unset()

    pawnlib_timeout = pawn.to_dict().get('PAWN_TIMEOUT', 10)
    if pawnlib_timeout > 0:
        pawnlib_timeout = pawnlib_timeout / 1000

    timeout = timeout or pawnlib_timeout

    url = append_http(url)
    (json_response, data, http_version, r_headers, error) = ({}, {}, None, None, None)

    response = None
    if method not in ("get", "post", "patch", "delete"):
        # cprint(f"[ERROR] unsupported method={method}, url={url} ", color="red")
        pawn.error_logger.error(f"unsupported method={method}, url={url} ") if pawn.error_logger else False
        return {"error": "unsupported method"}
    try:
        func = getattr(requests, method)
        if method == "get":
            response = func(url, verify=False, timeout=timeout, **kwargs)
        else:
            response = func(url, json=payload, verify=False, timeout=timeout, **kwargs)
        http_version = response.raw.version
        r_headers = response.headers

    except requests.exceptions.HTTPError as errh:
        error = errh
        if global_verbose > 0:
            kvPrint("Http Error:", errh)
        pawn.error_logger.error(f"Http Error:{errh}") if pawn.error_logger else False

    except requests.exceptions.ConnectionError as errc:
        error = errc
        if ("[Errno 11001] getaddrinfo failed" in str(errc) or  # Windows
                "[Errno -2] Name or service not known" in str(errc) or  # Linux
                "[Errno 8] nodename nor servname " in str(errc)):  # OS X
            errc = "DNSLookupError"
        if global_verbose > 0:
            kvPrint("Error Connecting:", errc, "FAIL")
        pawn.error_logger.error(f"Error Connecting:{errc}, {url}") if pawn.error_logger else False

    except requests.exceptions.Timeout as errt:
        error = errt
        if global_verbose > 0:
            kvPrint("Timeout Error:", errt, "FAIL")
        pawn.error_logger.error(f"Timeout Connecting:{errt}, {url}") if pawn.error_logger else False

    except requests.exceptions.RequestException as err:
        error = err
        if global_verbose > 0:
            kvPrint("OOps: Something Else", err, "FAIL")
        pawn.error_logger.error(f"OOps: Something Else:{err}, {url}") if pawn.error_logger else False

    # cprint(f"----> {url}, {method}, {payload} , {response.status_code}", "green")

    try:
        response_code = response.status_code
    except:
        response_code = 999

    json_payload = json.dumps(payload)
    if global_verbose > 1:
        debug_logging(f"{url}, {method}, {json_payload} , {response_code}")

    if response_code != 999:
        try:
            json_response = response.json()
        except:
            json_response = {}
            data["text"] = response.text
            # debug_logging(f"{url} , response is not json -> '{response.text}'")

        if elapsed:
            data["elapsed"] = int(response.elapsed.total_seconds() * 1000)
            if len(json_response) > 0:
                json_response["elapsed"] = data["elapsed"]

    if response_code > 200 and response_code != 999 and print_error:
        text = response.text
        pawn.error_logger.error(f"status_code: {response_code} , url: {url} , payload: {payload}, response: {text}") if pawn.error_logger else False
    data["status_code"] = response_code
    data["http_version"] = http_version
    data["r_headers"] = r_headers
    data["json"] = json_response
    data["error"] = error
    if print_error:
        if error:
            pawn.error_logger.error(f"{error}") if pawn.error_logger else False

    if isinstance(payload, dict):
        payload = json.dumps(payload)

    pawn.console.debug(f"url={url}, method={method}, payload={payload}, kwargs={kwargs}, error={error}")

    return data


icon_rpc_call = IconRpcHelper().rpc_call



#!/usr/bin/env python3
import argparse
from pawnlib.builder.generator import generate_banner
from pawnlib.__version__ import __version__ as _version
from pawnlib.config import pawn, pconf
import os
import json
from pawnlib.typing import str2bool, StackList, remove_tags
from pawnlib.output import write_json
from rich.tree import Tree
from pawnlib.resource import get_interface_ips, get_public_ip, get_hostname, get_platform_info, get_rlimit_nofile, get_mem_info, get_ip_and_netmask

__description__ = "This command displays server resource information."


def get_parser():
    parser = argparse.ArgumentParser(description='monitor')
    parser = get_arguments(parser)
    return parser


def get_arguments(parser):
    parser.add_argument('url', help='url', type=str, nargs='?', default="")
    parser.add_argument('-c', '--config-file', type=str, help='config', default="config.ini")
    parser.add_argument('-v', '--verbose', action='count', help='verbose mode. view level (default: %(default)s)', default=1)
    parser.add_argument('-q', '--quiet', action='count', help='Quiet mode. Dont show any messages. (default: %(default)s)', default=0)
    parser.add_argument('-b', '--base-dir', type=str, help='base dir for httping (default: %(default)s)', default=os.getcwd())
    parser.add_argument("--write-filename", '-w', type=str, help="write filename", default="")
    return parser


def print_banner():
    banner = generate_banner(
        app_name=pconf().app_name,
        author="jinwoo",
        description=f"{__description__} \n"
                    f" - base_dir    : {pconf().args.base_dir} \n" 
                    f" - logs_dir    : {pconf().args.base_dir}/logs \n",
        font="graffiti",
        version=_version
    )
    print(banner)


def main():
    app_name = 'Resource Information'
    parser = get_parser()
    args, unknown = parser.parse_known_args()
    config_file = args.config_file

    is_hide_line_number = args.verbose > 2
    stdout = not args.quiet

    pawn.set(
        PAWN_CONFIG_FILE=config_file,
        PAWN_PATH=args.base_dir,
        PAWN_CONSOLE=dict(
            redirect=True,
            record=True,
            log_path=is_hide_line_number, # hide line number on the right side
        ),
        app_name=app_name,
        args=args,
        try_pass=False,
        last_execute_point=0,
        data={
            "response_time": StackList(),
        },
        fail_count=0,
        total_count=0,

    )
    if args.verbose > 2:
        pawn.set(
            PAWN_LOGGER=dict(
                log_level="DEBUG",
                stdout_level="DEBUG",
            )
        )
    print_banner()

    result = {
        "system": {},
        "network": {},
    }

    system_tree = Tree("[bold]🖥️  System Information[/bold]")
    system_tree.add(f"Hostname:  {get_hostname()}")
    for k, v in get_platform_info().items():
        system_tree.add(f"{k.title()}: {v}")
        result['system'][k] = v

    result['system']['mem_total'] = get_mem_info().get('mem_total')
    system_tree.add(f"Memory: {result['system']['mem_total']} GB")

    result['system']["resource_limit"] = get_rlimit_nofile()
    system_tree.add(f"Resource limit: {result['system']['resource_limit']}")

    pawn.console.print(system_tree)
    print()

    network_tree = Tree("[bold]🛜 Network Interface[/bold]")
    result['network']['public_ip'] = get_public_ip()
    network_tree.add(f"[bold] Public IP[/bold]: {result['network']['public_ip']}")

    local_tree = network_tree.add("[bold] Local IP[/bold]")

    interface_list = get_interface_ips(ignore_interfaces=['lo0', 'lo'])

    if interface_list:
        longest_length = max(len(item[0]) for item in interface_list)

        for interface, ipaddr in get_interface_ips(ignore_interfaces=['lo0', 'lo']):
            result['network'][interface] = ipaddr
            if "G/W" in ipaddr:
                interface = f"[bold blue][on #050B27]{interface:<{longest_length}} [/bold blue]"
                ipaddr = f"{ipaddr}[/on #050B27]"
            local_tree.add(f"[bold]{interface:<{longest_length+1}}[/bold]: {ipaddr}")
        pawn.console.print(network_tree)

        if args.write_filename:
            write_res = write_json(filename=args.write_filename, data=result)
            pawn.console.log(write_res)


if __name__ == '__main__':
    try:
        main()
    except Exception as e:
        pawn.console.log(e)

